# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

from unittest import mock

import pytest

import tanjun


class TestHooks:
    def test_add_to_command(self):
        hooks = tanjun.AnyHooks()
        mock_command = mock.Mock()

        assert hooks.add_to_command(mock_command) is mock_command

        mock_command.set_hooks.assert_called_once_with(hooks)

    @pytest.mark.skip(reason="not implemented")
    def test_copy(self):
        hooks = tanjun.AnyHooks()

        result = hooks.copy()

        assert result == hooks
        assert result is not hooks

    def test_add_on_error(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().set_on_error(mock_other_callback)
        mock_on_error = mock.Mock()

        assert hooks.add_on_error(mock_on_error) is hooks

        assert hooks._error_callbacks == [mock_other_callback, mock_on_error]

    def test_set_on_error(self):
        hooks = tanjun.AnyHooks().set_on_error(mock.Mock())
        mock_on_error = mock.Mock()

        assert hooks.set_on_error(mock_on_error) is hooks

        assert hooks._error_callbacks == [mock_on_error]

    def test_set_on_error_when_none(self):
        hooks = tanjun.AnyHooks().set_on_error(mock.Mock())

        assert hooks.set_on_error(None) is hooks

        assert hooks._error_callbacks == []

    def test_with_on_error(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().set_on_error(mock_other_callback)
        mock_on_error = mock.Mock()

        assert hooks.with_on_error(mock_on_error) is mock_on_error

        assert hooks._error_callbacks == [mock_other_callback, mock_on_error]

    def test_add_on_parser_error(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_on_parser_error(mock_other_callback)
        mock_on_parser_error = mock.Mock()

        assert hooks.add_on_parser_error(mock_on_parser_error) is hooks

        assert hooks._parser_error_callbacks == [mock_other_callback, mock_on_parser_error]

    def test_set_on_parser_error(self):
        hooks = tanjun.AnyHooks().add_on_parser_error(mock.Mock())
        mock_on_parser_error = mock.Mock()

        assert hooks.set_on_parser_error(mock_on_parser_error) is hooks

        assert hooks._parser_error_callbacks == [mock_on_parser_error]

    def test_set_on_parser_error_when_none(self):
        hooks = tanjun.AnyHooks().add_on_parser_error(mock.Mock())

        assert hooks.set_on_parser_error(None) is hooks

        assert hooks._parser_error_callbacks == []

    def test_with_on_parser_error(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_on_parser_error(mock_other_callback)
        mock_on_parser_error = mock.Mock()

        assert hooks.with_on_parser_error(mock_on_parser_error) is mock_on_parser_error

        assert hooks._parser_error_callbacks == [mock_other_callback, mock_on_parser_error]

    def test_add_post_execution(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().set_post_execution(mock_other_callback)
        mock_post_execution = mock.Mock()

        assert hooks.add_post_execution(mock_post_execution) is hooks

        assert hooks._post_execution_callbacks == [mock_other_callback, mock_post_execution]

    def test_set_post_execution(self):
        hooks = tanjun.AnyHooks().set_post_execution(mock.Mock())
        mock_post_execution = mock.Mock()

        assert hooks.set_post_execution(mock_post_execution) is hooks

        assert hooks._post_execution_callbacks == [mock_post_execution]

    def test_set_post_execution_when_none(self):
        hooks = tanjun.AnyHooks().set_post_execution(mock.Mock())

        assert hooks.set_post_execution(None) is hooks

        assert hooks._post_execution_callbacks == []

    def test_with_post_execution(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_post_execution(mock_other_callback)
        mock_post_execution = mock.Mock()

        assert hooks.with_post_execution(mock_post_execution) is mock_post_execution

        assert hooks._post_execution_callbacks == [mock_other_callback, mock_post_execution]

    def test_add_pre_execution(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_pre_execution(mock_other_callback)
        mock_pre_execution = mock.Mock()

        assert hooks.add_pre_execution(mock_pre_execution) is hooks

        assert hooks._pre_execution_callbacks == [mock_other_callback, mock_pre_execution]

    def test_set_pre_execution(self):
        hooks = tanjun.AnyHooks().add_pre_execution(mock.Mock())
        mock_pre_execution = mock.Mock()

        assert hooks.set_pre_execution(mock_pre_execution) is hooks

        assert hooks._pre_execution_callbacks == [mock_pre_execution]

    def test_set_pre_execution_when_none(self):
        hooks = tanjun.AnyHooks().add_pre_execution(mock.Mock())

        assert hooks.set_pre_execution(None) is hooks

        assert hooks._pre_execution_callbacks == []

    def test_with_pre_execution(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_pre_execution(mock_other_callback)
        mock_pre_execution = mock.Mock()

        assert hooks.with_pre_execution(mock_pre_execution) is mock_pre_execution

        assert hooks._pre_execution_callbacks == [mock_other_callback, mock_pre_execution]

    def test_add_on_success(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_on_success(mock_other_callback)
        mock_on_success = mock.Mock()

        assert hooks.add_on_success(mock_on_success) is hooks

        assert hooks._success_callbacks == [mock_other_callback, mock_on_success]

    def test_set_on_success(self):
        hooks = tanjun.AnyHooks().add_on_success(mock.Mock())
        mock_on_success = mock.Mock()

        assert hooks.set_on_success(mock_on_success) is hooks

        assert hooks._success_callbacks == [mock_on_success]

    def test_set_on_success_when_none(self):
        hooks = tanjun.AnyHooks().add_on_success(mock.Mock())

        assert hooks.set_on_success(None) is hooks

        assert hooks._success_callbacks == []

    def test_with_on_success(self):
        mock_other_callback = mock.Mock()
        hooks = tanjun.AnyHooks().add_on_success(mock_other_callback)
        mock_on_success = mock.Mock()

        assert hooks.with_on_success(mock_on_success) is mock_on_success

        assert hooks._success_callbacks == [mock_other_callback, mock_on_success]

    @pytest.mark.asyncio()
    async def test_trigger_error_for_parser_error_with_handlers(self):
        mock_callback = mock.Mock()
        mock_other_hook = mock.Mock(trigger_error=mock.AsyncMock(return_value=100))
        mock_context = mock.AsyncMock()
        mock_error = mock.MagicMock(tanjun.ParserError)

        result = (
            await tanjun.AnyHooks()
            .set_on_parser_error(mock_callback)
            .trigger_error(mock_context, mock_error, hooks={mock_other_hook})
        )

        assert result == 200
        mock_context.call_with_async_di.assert_awaited_once_with(mock_callback, mock_context, mock_error)
        mock_other_hook.trigger_error.assert_awaited_once_with(mock_context, mock_error)

    @pytest.mark.asyncio()
    async def test_trigger_error_for_parser_error_without_handlers(self):
        result = await tanjun.AnyHooks().trigger_error(mock.Mock(), mock.MagicMock(tanjun.ParserError))

        assert result == 0

    @pytest.mark.asyncio()
    async def test_trigger_error_with_handler(self):
        mock_callback = mock.Mock()
        mock_other_hook = mock.Mock(trigger_error=mock.AsyncMock(return_value=2))
        mock_context = mock.AsyncMock()
        mock_context.call_with_async_di.return_value = True
        mock_error = mock.Mock()

        result = (
            await tanjun.AnyHooks()
            .set_on_error(mock_callback)
            .trigger_error(mock_context, mock_error, hooks={mock_other_hook})
        )

        assert result == 3
        mock_context.call_with_async_di.assert_awaited_once_with(mock_callback, mock_context, mock_error)
        mock_other_hook.trigger_error.assert_awaited_once_with(mock_context, mock_error)

    @pytest.mark.asyncio()
    async def test_trigger_error_without_handler(self):
        result = await tanjun.AnyHooks().trigger_error(mock.Mock(), mock.Mock())

        assert result == 0

    @pytest.mark.asyncio()
    async def test_trigger_post_execution_with_handlers(self):
        mock_callback = mock.Mock()
        mock_other_hook = mock.Mock(trigger_post_execution=mock.AsyncMock())
        mock_context = mock.AsyncMock()

        (
            await tanjun.AnyHooks()
            .set_post_execution(mock_callback)
            .trigger_post_execution(mock_context, hooks={mock_other_hook})
        )

        mock_context.call_with_async_di.assert_awaited_once_with(mock_callback, mock_context)
        mock_other_hook.trigger_post_execution.assert_awaited_once_with(mock_context)

    @pytest.mark.asyncio()
    async def test_trigger_post_execution_without_handlers(self):
        await tanjun.AnyHooks().trigger_post_execution(mock.Mock())

    @pytest.mark.asyncio()
    async def test_trigger_pre_execution_with_handlers(self):
        mock_callback = mock.Mock()
        mock_other_hook = mock.Mock(trigger_pre_execution=mock.AsyncMock())
        mock_context = mock.AsyncMock()

        (
            await tanjun.AnyHooks()
            .set_pre_execution(mock_callback)
            .trigger_pre_execution(mock_context, hooks={mock_other_hook})
        )

        mock_context.call_with_async_di.assert_awaited_once_with(mock_callback, mock_context)
        mock_other_hook.trigger_pre_execution.assert_awaited_once_with(mock_context)

    @pytest.mark.asyncio()
    async def test_trigger_pre_execution_without_handlers(self):
        await tanjun.AnyHooks().trigger_pre_execution(mock.Mock())

    @pytest.mark.asyncio()
    async def test_trigger_success_with_handlers(self):
        mock_callback = mock.Mock()
        mock_other_hook = mock.Mock(trigger_success=mock.AsyncMock())
        mock_context = mock.AsyncMock()

        (await tanjun.AnyHooks().set_on_success(mock_callback).trigger_success(mock_context, hooks={mock_other_hook}))

        mock_context.call_with_async_di.assert_awaited_once_with(mock_callback, mock_context)
        mock_other_hook.trigger_success.assert_awaited_once_with(mock_context)

    @pytest.mark.asyncio()
    async def test_trigger_success_without_handlers(self):
        await tanjun.AnyHooks().trigger_success(mock.Mock())



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

import asyncio
import datetime
import types
import typing
from unittest import mock

import alluka
import hikari
import pytest

import tanjun

_T = typing.TypeVar("_T")


def stub_class(cls: type[_T], /, **namespace: typing.Any) -> type[_T]:
    namespace["__slots__"] = ()

    for name in getattr(cls, "__abstractmethods__", None) or ():
        if name not in namespace:
            namespace[name] = mock.MagicMock()

    name = origin.__name__ if (origin := getattr(cls, "__origin__", None)) else cls.__name__
    new_cls = types.new_class(name, (cls,), exec_body=lambda body: body.update(namespace))
    return typing.cast(type[_T], new_cls)


@pytest.fixture()
def mock_client() -> tanjun.abc.Client:
    return mock.MagicMock(tanjun.abc.Client, rest=mock.AsyncMock(hikari.api.RESTClient))


@pytest.fixture()
def mock_component() -> tanjun.abc.Component:
    return mock.MagicMock(tanjun.abc.Component)


class TestMessageContext:
    @pytest.fixture()
    def context(self, mock_client: mock.Mock) -> tanjun.context.MessageContext:
        return tanjun.context.MessageContext(
            mock_client,
            "hi there",
            mock.AsyncMock(),
            triggering_name="bonjour",
            triggering_prefix="bonhoven",
        )

    def test___repr__(self, context: tanjun.context.MessageContext):
        assert repr(context) == f"MessageContext <{context.message!r}, {context.command!r}>"

    def test_author_property(self, context: tanjun.context.MessageContext):
        assert context.author is context.message.author

    def test_channel_id_property(self, context: tanjun.context.MessageContext):
        assert context.channel_id is context.message.channel_id

    def test_created_at_property(self, context: tanjun.context.MessageContext):
        assert context.created_at is context.message.created_at

    def test_guild_id_property(self, context: tanjun.context.MessageContext):
        assert context.guild_id is context.message.guild_id

    def test_has_responded_property(self, context: tanjun.context.MessageContext):
        assert context.has_responded is False

    def test_has_responded_property_when_initial_repsonse_id_set(self, context: tanjun.context.MessageContext):
        context._initial_response_id = hikari.Snowflake(321123)

        assert context.has_responded is True

    def test_is_human_property(self, context: tanjun.context.MessageContext):
        context.message.author = mock.Mock(is_bot=False)
        context.message.webhook_id = None

        assert context.is_human is True

    def test_is_human_property_when_is_bot(self, context: tanjun.context.MessageContext):
        context.message.author = mock.Mock(is_bot=True)
        context.message.webhook_id = None

        assert context.is_human is False

    def test_is_human_property_when_is_webhook(self, context: tanjun.context.MessageContext):
        context.message.author = mock.Mock(is_bot=False)
        context.message.webhook_id = hikari.Snowflake(123321)

        assert context.is_human is False

    def test_member_property(self, context: tanjun.context.MessageContext):
        assert context.member is context.message.member

    def test_message_property(self, context: tanjun.context.MessageContext):
        assert context.message is context._message

    def test_set_command(self, context: tanjun.context.MessageContext):
        mock_command = mock.Mock()

        assert context.set_command(mock_command) is context

        assert context.command is mock_command
        assert context.get_type_dependency(tanjun.abc.ExecutableCommand) is mock_command
        assert context.get_type_dependency(tanjun.abc.MessageCommand) is mock_command

    def test_set_command_when_none(self, context: tanjun.context.MessageContext):
        assert isinstance(context.client.injector.get_type_dependency, mock.Mock)
        context.client.injector.get_type_dependency.return_value = alluka.abc.UNDEFINED
        context.set_command(None)
        context.set_command(None)

        assert context.command is None
        assert context.get_type_dependency(tanjun.abc.ExecutableCommand) is alluka.abc.UNDEFINED
        assert context.get_type_dependency(tanjun.abc.MessageCommand) is alluka.abc.UNDEFINED

    def test_set_command_when_none_and_previously_set(self, context: tanjun.context.MessageContext):
        assert isinstance(context.client.injector.get_type_dependency, mock.Mock)
        context.client.injector.get_type_dependency.return_value = alluka.abc.UNDEFINED
        mock_command = mock.Mock()
        context.set_command(mock_command)
        context.set_command(None)

        assert context.command is None
        assert context.get_type_dependency(tanjun.abc.ExecutableCommand) is alluka.abc.UNDEFINED
        assert context.get_type_dependency(tanjun.abc.MessageCommand) is alluka.abc.UNDEFINED

    def test_set_command_when_finalised(self, context: tanjun.context.MessageContext):
        context.finalise()
        mock_command = mock.Mock()

        with pytest.raises(TypeError):
            context.set_command(mock_command)

        assert context.command is not mock_command

    def test_set_content(self, context: tanjun.context.MessageContext):
        assert context.set_content("hi") is context
        assert context.content == "hi"

    def test_set_content_when_finalised(self, context: tanjun.context.MessageContext):
        context.finalise()

        with pytest.raises(TypeError):
            context.set_content("hi")

        assert context.content != "hi"

    def test_set_triggering_name(self, context: tanjun.context.MessageContext):
        assert context.set_triggering_name("bonjour") is context

        assert context.triggering_name == "bonjour"

    def test_set_triggering_name_when_finalised(self, context: tanjun.context.MessageContext):
        context.finalise()

        with pytest.raises(TypeError):
            context.set_triggering_name("bonjour2")

        assert context.triggering_name != "bonjour2"

    def test_set_triggering_prefix(self, context: tanjun.context.MessageContext):
        assert context.set_triggering_prefix("bonhoven") is context

        assert context.triggering_prefix == "bonhoven"

    def test_set_triggering_prefix_when_finalised(self, context: tanjun.context.MessageContext):
        context.finalise()

        with pytest.raises(TypeError):
            context.set_triggering_prefix("bonhoven2")

        assert context.triggering_prefix != "bonhoven2"

    @pytest.mark.asyncio()
    async def test_delete_initial_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._initial_response_id = hikari.Snowflake(32123)

        await context.delete_initial_response()

        mock_client.rest.delete_message.assert_awaited_once_with(context.message.channel_id, 32123)

    @pytest.mark.asyncio()
    async def test_delete_initial_response_when_no_initial_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        with pytest.raises(LookupError):
            await context.delete_initial_response()

        mock_client.rest.delete_message.assert_not_called()

    @pytest.mark.asyncio()
    async def test_delete_last_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._last_response_id = hikari.Snowflake(32123)

        await context.delete_last_response()

        mock_client.rest.delete_message.assert_awaited_once_with(context.message.channel_id, 32123)

    @pytest.mark.asyncio()
    async def test_delete_last_response_when_no_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        with pytest.raises(LookupError):
            await context.delete_last_response()

        mock_client.rest.delete_message.assert_not_called()

    @pytest.mark.asyncio()
    async def test_edit_initial_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._initial_response_id = hikari.Snowflake(32123)
        mock_attachment = mock.Mock()
        mock_attachments = [mock.Mock()]
        mock_component = mock.Mock()
        mock_components = [mock.Mock()]
        mock_embed = mock.Mock()
        mock_embeds = [mock.Mock()]

        with mock.patch.object(asyncio, "create_task") as create_task:
            await context.edit_initial_response(
                "hi",
                attachment=mock_attachment,
                attachments=mock_attachments,
                component=mock_component,
                components=mock_components,
                embed=mock_embed,
                embeds=mock_embeds,
                replace_attachments=True,
                mentions_everyone=False,
                user_mentions=[123, 321],
                role_mentions=[321243],
            )

        mock_client.rest.edit_message.assert_awaited_once_with(
            context.message.channel_id,
            32123,
            content="hi",
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            replace_attachments=True,
            mentions_everyone=False,
            user_mentions=[123, 321],
            role_mentions=[321243],
        )
        create_task.assert_not_called()

    @pytest.mark.asyncio()
    async def test_edit_initial_response_when_no_initial_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        with pytest.raises(LookupError):
            await context.edit_initial_response("hi")

        mock_client.rest.edit_message.assert_not_called()

    @pytest.mark.parametrize("delete_after", [datetime.timedelta(seconds=123), 123, 123.0])
    @pytest.mark.asyncio()
    async def test_edit_initial_response_when_delete_after(
        self,
        delete_after: typing.Union[datetime.timedelta, float, int],
        context: tanjun.context.MessageContext,
        mock_client: mock.Mock,
    ):
        mock_delete_after = mock.Mock()
        context = stub_class(tanjun.context.MessageContext, _delete_after=mock_delete_after)(
            mock_client, "e", mock.AsyncMock()
        )
        context._initial_response_id = hikari.Snowflake(32123)

        with mock.patch.object(asyncio, "create_task") as create_task:
            await context.edit_initial_response("hi", delete_after=delete_after)

        create_task.assert_called_once_with(mock_delete_after.return_value)
        mock_delete_after.assert_called_once_with(123.0, mock_client.rest.edit_message.return_value)

    @pytest.mark.asyncio()
    async def test_edit_last_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._last_response_id = hikari.Snowflake(32123)
        mock_attachment = mock.Mock()
        mock_attachments = [mock.Mock()]
        mock_component = mock.Mock()
        mock_components = [mock.Mock()]
        mock_embed = mock.Mock()
        mock_embeds = [mock.Mock()]

        await context.edit_last_response(
            "hi",
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            replace_attachments=True,
            mentions_everyone=False,
            user_mentions=[123, 321],
            role_mentions=[321243],
        )

        mock_client.rest.edit_message.assert_awaited_once_with(
            context.message.channel_id,
            32123,
            content="hi",
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            replace_attachments=True,
            mentions_everyone=False,
            user_mentions=[123, 321],
            role_mentions=[321243],
        )

    @pytest.mark.asyncio()
    async def test_edit_last_response_when_no_last_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        with pytest.raises(LookupError):
            await context.edit_last_response("hi")

        mock_client.rest.edit_message.assert_not_called()

    @pytest.mark.parametrize("delete_after", [datetime.timedelta(seconds=654), 654, 654.0])
    @pytest.mark.asyncio()
    async def test_edit_last_response_when_delete_after(
        self,
        context: tanjun.context.MessageContext,
        mock_client: mock.Mock,
        delete_after: typing.Union[datetime.timedelta, int, float],
    ):
        mock_delete_after = mock.Mock()
        context = stub_class(tanjun.context.MessageContext, _delete_after=mock_delete_after)(
            mock_client, "e", mock.AsyncMock()
        )
        context._last_response_id = hikari.Snowflake(32123)

        with mock.patch.object(asyncio, "create_task") as create_task:
            await context.edit_last_response("hi", delete_after=delete_after)

        create_task.assert_called_once_with(mock_delete_after.return_value)
        mock_delete_after.assert_called_once_with(654.0, mock_client.rest.edit_message.return_value)

    @pytest.mark.asyncio()
    async def test_fetch_initial_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._initial_response_id = hikari.Snowflake(32123)

        message = await context.fetch_initial_response()

        assert message is mock_client.rest.fetch_message.return_value
        mock_client.rest.fetch_message.assert_awaited_once_with(context.message.channel_id, 32123)

    @pytest.mark.asyncio()
    async def test_fetch_initial_response_when_no_initial_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        with pytest.raises(LookupError):
            await context.fetch_initial_response()

        mock_client.rest.fetch_message.assert_not_called()

    @pytest.mark.asyncio()
    async def test_fetch_last_response(self, context: tanjun.context.MessageContext, mock_client: mock.Mock):
        context._last_response_id = hikari.Snowflake(32123)

        message = await context.fetch_last_response()

        assert message is mock_client.rest.fetch_message.return_value
        mock_client.rest.fetch_message.assert_awaited_once_with(context.message.channel_id, 32123)

    @pytest.mark.asyncio()
    async def test_fetch_last_response_when_no_last_response(
        self, context: tanjun.context.MessageContext, mock_client: mock.Mock
    ):
        context._last_response_id = None
        with pytest.raises(LookupError):
            await context.fetch_last_response()

        mock_client.rest.fetch_message.assert_not_called()

    @pytest.mark.asyncio()
    async def test__delete_after(self, context: tanjun.context.MessageContext):
        mock_message = mock.AsyncMock()

        with mock.patch.object(asyncio, "sleep") as sleep:
            await context._delete_after(1545.4, mock_message)

            sleep.assert_awaited_once_with(1545.4)
            mock_message.delete.assert_awaited_once_with()

    @pytest.mark.asyncio()
    async def test__delete_after_handles_not_found_error(self, context: tanjun.context.MessageContext):
        mock_message = mock.AsyncMock()
        mock_message.delete.side_effect = hikari.NotFoundError(url="", headers={}, raw_body=None)

        with mock.patch.object(asyncio, "sleep") as sleep:
            await context._delete_after(1545.4, mock_message)

            sleep.assert_awaited_once_with(1545.4)
            mock_message.delete.assert_awaited_once_with()

    @pytest.mark.asyncio()
    async def test_respond(self, context: tanjun.context.MessageContext):
        mock_attachment = mock.Mock()
        mock_attachments = [mock.Mock()]
        mock_component = mock.Mock()
        mock_components = [mock.Mock()]
        mock_embed = mock.Mock()
        mock_embeds = [mock.Mock()]

        with mock.patch.object(asyncio, "create_task") as create_task:
            await context.respond(
                "hi",
                attachment=mock_attachment,
                attachments=mock_attachments,
                component=mock_component,
                components=mock_components,
                embed=mock_embed,
                embeds=mock_embeds,
                tts=True,
                nonce="nonce",
                reply=432123,
                mentions_everyone=False,
                mentions_reply=True,
                user_mentions=[123, 321],
                role_mentions=[555, 444],
            )

        create_task.assert_not_called()
        assert isinstance(context.message.respond, mock.Mock)
        context.message.respond.assert_awaited_once_with(
            content="hi",
            attachment=mock_attachment,
            attachments=mock_attachments,
            component=mock_component,
            components=mock_components,
            embed=mock_embed,
            embeds=mock_embeds,
            tts=True,
            nonce="nonce",
            reply=432123,
            mentions_everyone=False,
            mentions_reply=True,
            user_mentions=[123, 321],
            role_mentions=[555, 444],
        )
        assert context._last_response_id == context.message.respond.return_value.id
        assert context._initial_response_id == context.message.respond.return_value.id

    @pytest.mark.asyncio()
    async def test_respond_when_initial_response_id_already_set(self, context: tanjun.context.MessageContext):
        context._initial_response_id = hikari.Snowflake(32123)

        await context.respond("hi")

        context._initial_response_id == 32123

    @pytest.mark.parametrize("delete_after", [datetime.timedelta(seconds=123), 123, 123.0])
    @pytest.mark.asyncio()
    async def test_respond_when_delete_after(self, delete_after: typing.Union[int, float, datetime.timedelta]):
        mock_delete_after = mock.Mock()
        context = stub_class(tanjun.context.MessageContext, _delete_after=mock_delete_after)(
            mock.Mock(), "e", mock.AsyncMock()
        )

        with mock.patch.object(asyncio, "create_task") as create_task:
            await context.respond("hi", delete_after=delete_after)

        assert isinstance(context.message.respond, mock.Mock)
        mock_delete_after.assert_called_once_with(123.0, context.message.respond.return_value)
        create_task.assert_called_once_with(mock_delete_after.return_value)



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Functions and classes used to enable more Discord oriented argument converters."""
from __future__ import annotations

__all__: list[str] = [
    "ToChannel",
    "ToEmoji",
    "ToGuild",
    "ToInvite",
    "ToInviteWithMetadata",
    "ToMember",
    "ToMessage",
    "ToPresence",
    "ToRole",
    "ToUser",
    "ToVoiceState",
    "from_datetime",
    "parse_channel_id",
    "parse_emoji_id",
    "parse_message_id",
    "parse_role_id",
    "parse_snowflake",
    "parse_user_id",
    "search_channel_ids",
    "search_emoji_ids",
    "search_role_ids",
    "search_snowflakes",
    "search_user_ids",
    "to_bool",
    "to_channel",
    "to_color",
    "to_colour",
    "to_datetime",
    "to_emoji",
    "to_guild",
    "to_invite",
    "to_invite_with_metadata",
    "to_member",
    "to_message",
    "to_presence",
    "to_role",
    "to_snowflake",
    "to_user",
    "to_voice_state",
]

import abc
import datetime
import logging
import operator
import re
import typing
import urllib.parse as urlparse
from collections import abc as collections

import alluka
import hikari

from . import abc as tanjun_abc
from .dependencies import async_cache

if typing.TYPE_CHECKING:
    from . import parsing


_ArgumentT = typing.Union[str, int, float]
_ValueT = typing.TypeVar("_ValueT")
_LOGGER = logging.getLogger("hikari.tanjun.conversion")


class BaseConverter(typing.Generic[_ValueT], abc.ABC):
    """Base class for the standard converters.

    .. warning::
        Inheriting from this is completely unnecessary and should be avoided
        for people using the library unless they know what they're doing.


    This is detail of the standard implementation and isn't guaranteed to work
    between implementations but will work for implementations which provide
    the standard dependency injection or special cased support for these.

    While it isn't necessary to subclass this to implement your own converters
    since dependency injection can be used to access fields like the current Context,
    this class introduces some niceties around stuff like state warnings.
    """

    __slots__ = ("__weakref__",)
    __pdoc__: typing.ClassVar[dict[str, bool]] = {
        "async_cache": False,
        "cache_components": False,
        "intents": False,
        "requires_cache": False,
        "__pdoc__": False,
    }

    @property
    @abc.abstractmethod
    def async_caches(self) -> collections.Sequence[typing.Any]:
        """Collection of the asynchronous caches that this converter relies on.

        This will only be necessary if the suggested intents or cache_components
        aren't enabled for a converter which requires cache.
        """

    @property
    @abc.abstractmethod
    def cache_components(self) -> hikari.CacheComponents:
        """Cache component(s) the converter takes advantage of.

        .. note::
            Unless `BaseConverter.requires_cache` is `True`, these cache components
            aren't necessary but simply avoid the converter from falling back to
            REST requests.

        This will be `hikari.CacheComponents.NONE` if the converter doesn't
        make cache calls.
        """

    @property
    @abc.abstractmethod
    def intents(self) -> hikari.Intents:
        """Gateway intents this converter takes advantage of.

        .. note::
            This field is supplementary to `BaseConverter.cache_components` and
            is used to detect when the relevant component might not actually be
            being kept up to date or filled by gateway events.

            Unless `BaseConverter.requires_cache` is `True`, these intents being
            disabled won't stop this converter from working as it'll still fall
            back to REST requests.
        """

    @property
    @abc.abstractmethod
    def requires_cache(self) -> bool:
        """Whether this converter relies on the relevant cache stores to work.

        If this is `True` then this converter will not function properly
        in an environment `BaseConverter.intents` or `BaseConverter.cache_components`
        isn't satisfied and will never fallback to REST requests.
        """

    def check_client(self, client: tanjun_abc.Client, parent_name: str, /) -> None:
        """Check that this converter will work with the given client.

        This never raises any errors but simply warns the user if the converter
        is not compatible with the given client.

        Parameters
        ----------
        client : tanjun.abc.Client
            The client to check against.
        parent_name : str
            The name of the converter's parent, used for warning messages.
        """
        if not client.cache and any(
            client.get_type_dependency(cls) is alluka.abc.UNDEFINED for cls in self.async_caches
        ):
            if self.requires_cache:
                _LOGGER.warning(
                    f"Converter {self!r} registered with {parent_name} will always fail with a stateless client.",
                )

            elif self.cache_components:
                _LOGGER.warning(
                    f"Converter {self!r} registered with {parent_name} may not perform optimally in a stateless client.",
                )

        # elif missing_components := (self.cache_components & ~client.cache.components):
        #     _LOGGER.warning(

        if client.shards and (missing_intents := self.intents & ~client.shards.intents):
            _LOGGER.warning(
                f"Converter {self!r} registered with {parent_name} may not perform as expected "
                f"without the following intents: {missing_intents}",
            )


_DmCacheT = typing.Optional[async_cache.SfCache[hikari.DMChannel]]
_GuildChannelCacheT = typing.Optional[async_cache.SfCache[hikari.PartialChannel]]


# TODO: GuildChannelConverter
class ToChannel(BaseConverter[hikari.PartialChannel]):
    """Standard converter for channels mentions/IDs.

    For a standard instance of this see `to_channel`.
    """

    __slots__ = ("_include_dms",)

    def __init__(self, *, include_dms: bool = True) -> None:
        """Initialise a to channel converter.

        Other Parameters
        ----------------
        include_dms : bool
            Whether to include DM channels in the results.

            May lead to a lot of extra fallbacks to REST requests if
            the client doesn't have a registered async cache for DMs.

            Defaults to `True`.
        """
        self._include_dms = include_dms

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_GuildChannelCacheT, _DmCacheT)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.GUILD_CHANNELS

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _GuildChannelCacheT = alluka.inject(type=_GuildChannelCacheT),
        dm_cache: _DmCacheT = alluka.inject(type=_DmCacheT),
    ) -> hikari.PartialChannel:
        channel_id = parse_channel_id(argument, message="No valid channel mention or ID found")
        if ctx.cache and (channel_ := ctx.cache.get_guild_channel(channel_id)):
            return channel_

        no_guild_channel = False
        if cache:
            try:
                return await cache.get(channel_id)

            except async_cache.EntryNotFound:
                if not self._include_dms:
                    raise ValueError("Couldn't find channel") from None

                no_guild_channel = True

            except async_cache.CacheMissError:
                pass

        if dm_cache and self._include_dms:
            try:
                return await dm_cache.get(channel_id)

            except async_cache.EntryNotFound:
                if no_guild_channel:
                    raise ValueError("Couldn't find channel") from None

            except async_cache.CacheMissError:
                pass

        try:
            channel = await ctx.rest.fetch_channel(channel_id)
            if self._include_dms or isinstance(channel, hikari.GuildChannel):
                return channel

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find channel")


ChannelConverter = ToChannel
"""Deprecated alias of `ToChannel`."""

_EmojiCacheT = typing.Optional[async_cache.SfCache[hikari.KnownCustomEmoji]]


class ToEmoji(BaseConverter[hikari.KnownCustomEmoji]):
    """Standard converter for custom emojis.

    For a standard instance of this see `to_emoji`.

    .. note::
        If you just want to convert inpute to a `hikari.Emoji`, `hikari.CustomEmoji`
        or `hikari.UnicodeEmoji` without making any cache or REST calls then you
        can just use the relevant `hikari.Emoji.parse`, `hikari.CustomEmoji.parse`
        or `hikari.UnicodeEmoji.parse` methods.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_EmojiCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.EMOJIS

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_EMOJIS | hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _EmojiCacheT = alluka.inject(type=_EmojiCacheT),
    ) -> hikari.KnownCustomEmoji:
        emoji_id = parse_emoji_id(argument, message="No valid emoji or emoji ID found")

        if ctx.cache and (emoji := ctx.cache.get_emoji(emoji_id)):
            return emoji

        if cache:
            try:
                return await cache.get(emoji_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find emoji") from None

            except async_cache.CacheMissError:
                pass

        if ctx.guild_id:
            try:
                return await ctx.rest.fetch_emoji(ctx.guild_id, emoji_id)

            except hikari.NotFoundError:
                pass

        raise ValueError("Couldn't find emoji")


EmojiConverter = ToEmoji
"""Deprecated alias of `ToEmoji`."""


_GuildCacheT = typing.Optional[async_cache.SfCache[hikari.Guild]]


class ToGuild(BaseConverter[hikari.Guild]):
    """Stanard converter for guilds.

    For a standard instance of this see `to_guild`.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_GuildCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.GUILDS

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _GuildCacheT = alluka.inject(type=_GuildCacheT),
    ) -> hikari.Guild:
        guild_id = parse_snowflake(argument, message="No valid guild ID found")
        if ctx.cache and (guild := ctx.cache.get_guild(guild_id)):
            return guild

        if cache:
            try:
                return await cache.get(guild_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find guild") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_guild(guild_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find guild")


GuildConverter = ToGuild
"""Deprecated alias of `ToGuild`."""

_InviteCacheT = typing.Optional[async_cache.AsyncCache[str, hikari.InviteWithMetadata]]


class ToInvite(BaseConverter[hikari.Invite]):
    """Standard converter for invites."""

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_InviteCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.INVITES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_INVITES

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _InviteCacheT = alluka.inject(type=_InviteCacheT),
    ) -> hikari.Invite:
        if not isinstance(argument, str):
            raise ValueError(f"`{argument}` is not a valid invite code")

        if ctx.cache and (invite := ctx.cache.get_invite(argument)):
            return invite

        if cache:
            try:
                return await cache.get(argument)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find invite") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_invite(argument)
        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find invite")


InviteConverter = ToInvite
"""Deprecated alias of `ToInvite`."""


class ToInviteWithMetadata(BaseConverter[hikari.InviteWithMetadata]):
    """Standard converter for invites with metadata.

    For a standard instance of this see `to_invite_with_metadata`.

    .. note::
        Unlike `InviteConverter`, this converter is cache dependent.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_InviteCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.INVITES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_INVITES

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: typing.Optional[_InviteCacheT] = alluka.inject(type=_InviteCacheT),
    ) -> hikari.InviteWithMetadata:
        if not isinstance(argument, str):
            raise ValueError(f"`{argument}` is not a valid invite code")

        if ctx.cache and (invite := ctx.cache.get_invite(argument)):
            return invite

        if cache and (invite := await cache.get(argument)):
            return invite

        raise ValueError("Couldn't find invite")


InviteWithMetadataConverter = ToInviteWithMetadata
"""Deprecated alias of `ToInviteWithMetadata`."""


_MemberCacheT = typing.Optional[async_cache.SfGuildBound[hikari.Member]]


class ToMember(BaseConverter[hikari.Member]):
    """Standard converter for guild members.

    For a standard instance of this see `to_member`.

    This converter allows both mentions, raw IDs and partial usernames/nicknames
    and only works within a guild context.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_MemberCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.MEMBERS

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_MEMBERS | hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _MemberCacheT = alluka.inject(type=_MemberCacheT),
    ) -> hikari.Member:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a member from a DM channel")

        try:
            user_id = parse_user_id(argument, message="No valid user mention or ID found")

        except ValueError:
            if isinstance(argument, str):
                try:
                    return (await ctx.rest.search_members(ctx.guild_id, argument))[0]

                except (hikari.NotFoundError, IndexError):
                    pass

        else:
            if ctx.cache and (member := ctx.cache.get_member(ctx.guild_id, user_id)):
                return member

            if cache:
                try:
                    return await cache.get_from_guild(ctx.guild_id, user_id)

                except async_cache.EntryNotFound:
                    raise ValueError("Couldn't find member in this guild") from None

                except async_cache.CacheMissError:
                    pass

            try:
                return await ctx.rest.fetch_member(ctx.guild_id, user_id)

            except hikari.NotFoundError:
                pass

        raise ValueError("Couldn't find member in this guild")


MemberConverter = ToMember
"""Deprecated alias of `ToMember`."""

_PresenceCacheT = typing.Optional[async_cache.SfGuildBound[hikari.MemberPresence]]


class ToPresence(BaseConverter[hikari.MemberPresence]):
    """Standard converter for presences.

    For a standard instance of this see `to_presence`.

    This converter is cache dependent and only works in a guild context.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_PresenceCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.PRESENCES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_PRESENCES | hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _PresenceCacheT = alluka.inject(type=_PresenceCacheT),
    ) -> hikari.MemberPresence:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a presence from a DM channel")

        user_id = parse_user_id(argument, message="No valid member mention or ID found")
        if ctx.cache and (presence := ctx.cache.get_presence(ctx.guild_id, user_id)):
            return presence

        if cache and (presence := await cache.get_from_guild(ctx.guild_id, user_id, default=None)):
            return presence

        raise ValueError("Couldn't find presence in current guild")


PresenceConverter = ToPresence
"""Deprecated alias of `ToPresence`."""

_RoleCacheT = typing.Optional[async_cache.SfCache[hikari.Role]]


class ToRole(BaseConverter[hikari.Role]):
    """Standard converter for guild roles.

    For a standard instance of this see `to_role`.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_RoleCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.ROLES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _RoleCacheT = alluka.inject(type=_RoleCacheT),
    ) -> hikari.Role:
        role_id = parse_role_id(argument, message="No valid role mention or ID found")
        if ctx.cache and (role := ctx.cache.get_role(role_id)):
            return role

        if cache:
            try:
                return await cache.get(role_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find role") from None

            except async_cache.CacheMissError:
                pass

        if ctx.guild_id:
            for role in await ctx.rest.fetch_roles(ctx.guild_id):
                if role.id == role_id:
                    return role

        raise ValueError("Couldn't find role")


RoleConverter = ToRole
"""Deprecated alias of `ToRole`."""

_UserCacheT = typing.Optional[async_cache.SfCache[hikari.User]]


class ToUser(BaseConverter[hikari.User]):
    """Standard converter for users.

    For a standard instance of this see `to_user`.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_UserCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.NONE

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILDS | hikari.Intents.GUILD_MEMBERS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _UserCacheT = alluka.inject(type=_UserCacheT),
    ) -> hikari.User:
        # TODO: search by name if this is a guild context
        user_id = parse_user_id(argument, message="No valid user mention or ID found")
        if ctx.cache and (user := ctx.cache.get_user(user_id)):
            return user

        if cache:
            try:
                return await cache.get(user_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find user") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_user(user_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find user")


UserConverter = ToUser
"""Deprecated alias of `ToUser`."""


_MessageCacheT = typing.Optional[async_cache.SfCache[hikari.Message]]


class ToMessage(BaseConverter[hikari.Message]):
    """Standard converter for messages.

    For a standard instance of this see `to_message`.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_MessageCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.MESSAGES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_MESSAGES | hikari.Intents.DM_MESSAGES

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return False

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _MessageCacheT = alluka.inject(type=_MessageCacheT),
    ) -> hikari.Message:
        channel_id, message_id = parse_message_id(argument)
        if ctx.cache and (message := ctx.cache.get_message(message_id)):
            return message

        if cache:
            try:
                return await cache.get(message_id)

            except async_cache.EntryNotFound:
                raise ValueError("Couldn't find message") from None

            except async_cache.CacheMissError:
                pass

        try:
            return await ctx.rest.fetch_message(channel_id or ctx.channel_id, message_id)

        except hikari.NotFoundError:
            pass

        raise ValueError("Couldn't find message")


_VoiceStateCacheT = typing.Optional[async_cache.SfGuildBound[hikari.VoiceState]]


class ToVoiceState(BaseConverter[hikari.VoiceState]):
    """Standard converter for voice states.

    For a standard instance of this see `to_voice_state`.

    .. note::
        This converter is cache dependent and only works in a guild context.
    """

    __slots__ = ()

    @property
    def async_caches(self) -> collections.Sequence[typing.Any]:
        # <<inherited docstring from BaseConverter>>.
        return (_VoiceStateCacheT,)

    @property
    def cache_components(self) -> hikari.CacheComponents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.CacheComponents.VOICE_STATES

    @property
    def intents(self) -> hikari.Intents:
        # <<inherited docstring from BaseConverter>>.
        return hikari.Intents.GUILD_VOICE_STATES | hikari.Intents.GUILDS

    @property
    def requires_cache(self) -> bool:
        # <<inherited docstring from BaseConverter>>.
        return True

    async def __call__(
        self,
        argument: _ArgumentT,
        /,
        ctx: tanjun_abc.Context = alluka.inject(type=tanjun_abc.Context),
        cache: _VoiceStateCacheT = alluka.inject(type=_VoiceStateCacheT),
    ) -> hikari.VoiceState:
        if ctx.guild_id is None:
            raise ValueError("Cannot get a voice state from a DM channel")

        user_id = parse_user_id(argument, message="No valid user mention or ID found")

        if ctx.cache and (state := ctx.cache.get_voice_state(ctx.guild_id, user_id)):
            return state

        if cache and (state := await cache.get_from_guild(ctx.guild_id, user_id, default=None)):
            return state

        raise ValueError("Voice state couldn't be found for current guild")


VoiceStateConverter = ToVoiceState
"""Deprecated alias of `ToVoiceState`."""


class _IDMatcherSig(typing.Protocol):
    def __call__(self, value: _ArgumentT, /, *, message: str = "No valid mention or ID found") -> hikari.Snowflake:
        raise NotImplementedError


def _make_snowflake_parser(regex: re.Pattern[str], /) -> _IDMatcherSig:
    def parse(value: _ArgumentT, /, *, message: str = "No valid mention or ID found") -> hikari.Snowflake:
        """Parse a snowflake from a string or int value.

        .. note::
            This only allows the relevant entity's mention format if applicable.

        Parameters
        ----------
        value: str | int
            The value to parse (this argument can only be passed positionally).

        Other Parameters
        ----------------
        message: str
            The error message to raise if the value cannot be parsed.

        Returns
        -------
        hikari.Snowflake
            The parsed snowflake.

        Raises
        ------
        ValueError
            If the value cannot be parsed.
        """
        result: typing.Optional[hikari.Snowflake] = None
        if isinstance(value, str):
            if value.isdigit():
                result = hikari.Snowflake(value)

            else:
                capture = next(regex.finditer(value), None)
                result = hikari.Snowflake(capture.groups()[0]) if capture else None

        else:
            try:
                # Technically passing a float here is invalid (typing wise)
                # but we handle that by catching TypeError
                result = hikari.Snowflake(operator.index(typing.cast(int, value)))

            except (TypeError, ValueError):
                pass

        # We should also range check the provided ID.
        if result is not None and _range_check(result):
            return result

        raise ValueError(message) from None

    return parse


_IDSearcherSig = collections.Callable[[_ArgumentT], list[hikari.Snowflake]]


def _range_check(snowflake: hikari.Snowflake, /) -> bool:
    return snowflake.min() <= snowflake <= snowflake.max()


def _make_snowflake_searcher(regex: re.Pattern[str], /) -> _IDSearcherSig:
    def parse(value: _ArgumentT, /) -> list[hikari.Snowflake]:
        """Get the snowflakes in a string.

        .. note::
            This only allows the relevant entity's mention format if applicable.

        Parameters
        ----------
        value: str | int
            The value to parse (this argument can only be passed positionally).

        Returns
        -------
        list[hikari.Snowflake]
            List of the IDs found in the string.
        """
        if isinstance(value, str):
            if value.isdigit() and _range_check(result := hikari.Snowflake(value)):
                return [result]

            results = filter(
                _range_check, map(hikari.Snowflake, (match.groups()[0] for match in regex.finditer(value)))
            )
            return [*results, *filter(_range_check, map(hikari.Snowflake, filter(str.isdigit, value.split())))]

        try:
            # Technically passing a float here is invalid (typing wise)
            # but we handle that by catching TypeError
            result = hikari.Snowflake(operator.index(typing.cast(int, value)))

        except (TypeError, ValueError):
            return []

        if _range_check(result):
            return [result]

        return []

    return parse


_SNOWFLAKE_REGEX = re.compile(r"<[@&?!#a]{0,3}(?::\w+:)?(\d+)>")
parse_snowflake: _IDMatcherSig = _make_snowflake_parser(_SNOWFLAKE_REGEX)
"""Parse a snowflake from a string or int value.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Other Parameters
----------------
message: str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed snowflake.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_snowflakes: _IDSearcherSig = _make_snowflake_searcher(_SNOWFLAKE_REGEX)
"""Get the snowflakes in a string.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the snowflakes in the string.
"""

_CHANNEL_ID_REGEX = re.compile(r"<#(\d+)>")
parse_channel_id: _IDMatcherSig = _make_snowflake_parser(_CHANNEL_ID_REGEX)
"""Parse a channel ID from a string or int value.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Other Parameters
----------------
message: str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed channel ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_channel_ids: _IDSearcherSig = _make_snowflake_searcher(_CHANNEL_ID_REGEX)
"""Get the channel IDs in a string.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the channel IDs in the string.
"""

_EMOJI_ID_REGEX = re.compile(r"<a?:\w+:(\d+)>")
parse_emoji_id: _IDMatcherSig = _make_snowflake_parser(_EMOJI_ID_REGEX)
"""Parse an Emoji ID from a string or int value.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Other Parameters
----------------
message: str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed Emoji ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_emoji_ids: _IDSearcherSig = _make_snowflake_searcher(_EMOJI_ID_REGEX)
"""Get the emoji IDs in a string.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the emoji IDs in the string.
"""

_ROLE_ID_REGEX = re.compile(r"<@&(\d+)>")
parse_role_id: _IDMatcherSig = _make_snowflake_parser(_ROLE_ID_REGEX)
"""Parse a role ID from a string or int value.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Other Parameters
----------------
message: str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed role ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_role_ids: _IDSearcherSig = _make_snowflake_searcher(_ROLE_ID_REGEX)
"""Get the role IDs in a string.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the role IDs in the string.
"""

_USER_ID_REGEX = re.compile(r"<@!?(\d+)>")
parse_user_id: _IDMatcherSig = _make_snowflake_parser(_USER_ID_REGEX)
"""Parse a user ID from a string or int value.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Other Parameters
----------------
message: str
    The error message to raise if the value cannot be parsed.

Returns
-------
hikari.Snowflake
    The parsed user ID.

Raises
------
ValueError
    If the value cannot be parsed.
"""

search_user_ids: _IDSearcherSig = _make_snowflake_searcher(_USER_ID_REGEX)
"""Get the user IDs in a string.

Parameters
----------
value: str | int
    The value to parse (this argument can only be passed positionally).

Returns
-------
list[hikari.Snowflake]
    List of the user IDs in the string.
"""

_MESSAGE_LINK_REGEX = re.compile(r"(\d+|@me)/(\d+)/(\d+)")


def parse_message_id(
    value: _ArgumentT,
    /,
    *,
    message: str = "No valid message link or ID found",
) -> tuple[typing.Optional[hikari.Snowflake], hikari.Snowflake]:
    """Parse a user ID from a string or int value.

    Parameters
    ----------
    value: str | int
        The value to parse (this argument can only be passed positionally).

    Other Parameters
    ----------------
    message: str
        The error message to raise if the value cannot be parsed.

    Returns
    -------
    tuple[hikari.Snowflake | None, hikari.Snowflake]
        The parsed channel and message IDs.

    Raises
    ------
    ValueError
        If the value cannot be parsed.
    """
    channel_id: typing.Optional[hikari.Snowflake] = None
    message_id: typing.Optional[hikari.Snowflake] = None

    if isinstance(value, str):
        if value.isdigit():
            message_id = hikari.Snowflake(value)

        else:
            capture = next(_MESSAGE_LINK_REGEX.finditer(value), None)
            if capture:
                channel_id = hikari.Snowflake(capture[2])
                message_id = hikari.Snowflake(capture[3])

    else:
        try:
            # Technically passing a float here is invalid (typing wise)
            # but we handle that by catching TypeError
            message_id = hikari.Snowflake(operator.index(typing.cast(int, value)))

        except (TypeError, ValueError):
            pass

    # We should also range check the provided ID.
    if channel_id is not None and not _range_check(channel_id):
        raise ValueError(message) from None

    if message_id is not None and _range_check(message_id):
        return channel_id, message_id

    raise ValueError(message) from None


def _build_url_parser(callback: collections.Callable[[str], _ValueT], /) -> collections.Callable[[str], _ValueT]:
    def parse(value: str, /) -> _ValueT:
        """Convert an argument to a `urllib.parse` type.

        Parameters
        ----------
        value: str
            The value to parse (this argument can only be passed positionally).

        Returns
        -------
        _ValueT
            The parsed URL.

        Raises
        ------
        ValueError
            If the argument couldn't be parsed.
        """
        if value.startswith("<") and value.endswith(">"):
            value = value[1:-1]

        return callback(value)

    return parse


defragment_url: collections.Callable[[str], urlparse.DefragResult] = _build_url_parser(urlparse.urldefrag)
"""Convert an argument to a defragmented URL.

Parameters
----------
value: str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.DefragResult
    The parsed URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""

parse_url: collections.Callable[[str], urlparse.ParseResult] = _build_url_parser(urlparse.urlparse)
"""Convert an argument to a parsed URL.

Parameters
----------
value: str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.ParseResult
    The parsed URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""


split_url: collections.Callable[[str], urlparse.SplitResult] = _build_url_parser(urlparse.urlsplit)
"""Convert an argument to a split URL.

Parameters
----------
value: str
    The value to parse (this argument can only be passed positionally).

Returns
-------
urllib.parse.SplitResult
    The split URL.

Raises
------
ValueError
    If the argument couldn't be parsed.
"""

_DATETIME_REGEX = re.compile(r"<-?t:(\d+)(?::\w)?>")


def to_datetime(value: str, /) -> datetime.datetime:
    """Parse a datetime from Discord's datetime format.

    More information on this format can be found at
    https://discord.com/developers/docs/reference#message-formatting-timestamp-styles

    Parameters
    ----------
    value: str
        The value to parse.

    Returns
    -------
    datetime.datetime
        The parsed datetime.

    Raises
    ------
    ValueError
        If the value cannot be parsed.
    """
    try:
        timestamp = int(next(_DATETIME_REGEX.finditer(value)).groups()[0])

    except StopIteration:
        raise ValueError("Not a valid datetime") from None

    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


_VALID_DATETIME_STYLES = frozenset(("t", "T", "d", "D", "f", "F", "R"))


def from_datetime(value: datetime.datetime, /, *, style: str = "f") -> str:
    """Format a datetime as Discord's datetime format.

    More information on this format can be found at
    https://discord.com/developers/docs/reference#message-formatting-timestamp-styles

    Parameters
    ----------
    value: datetime.datetime
        The datetime to format.

    Other Parameters
    ----------------
    style: str
        The style to use.

        The valid styles can be found at
        https://discord.com/developers/docs/reference#message-formatting-formats
        and this defaults to `"f"`.

    Returns
    -------
    str
        The formatted datetime.

    Raises
    ------
    ValueError
        If the provided datetime is timezone naive.
        If an invalid style is provided.
    """
    if style not in _VALID_DATETIME_STYLES:
        raise ValueError(f"Invalid style: {style}")

    if value.tzinfo is None:
        raise ValueError("Cannot convert naive datetimes, please specify a timezone.")

    return f"<t:{round(value.timestamp())}:{style}>"


_YES_VALUES = frozenset(("y", "yes", "t", "true", "on", "1"))
_NO_VALUES = frozenset(("n", "no", "f", "false", "off", "0"))


def to_bool(value: str, /) -> bool:
    """Convert user string input into a boolean value.

    Parameters
    ----------
    value: str
        The value to convert.

    Returns
    -------
    bool
        The converted value.

    Raises
    ------
    ValueError
        If the value cannot be converted.
    """
    value = value.lower().strip()
    if value in _YES_VALUES:
        return True

    if value in _NO_VALUES:
        return False

    raise ValueError(f"Invalid bool value `{value}`")


def to_color(argument: _ArgumentT, /) -> hikari.Color:
    """Convert user input to a `hikari.colors.Color` object."""
    if isinstance(argument, str):
        values = argument.split(" ")
        if all(value.isdigit() for value in values):
            if len(values) == 1:
                return hikari.Color.of(int(values[0]))

            return hikari.Color.of(list(map(int, values)))

        if len(values) == 1:
            return hikari.Color.of(values[0])

        raise ValueError("Not a valid color representation")

    return hikari.Color.of(argument)


_TYPE_OVERRIDES: dict[collections.Callable[..., typing.Any], collections.Callable[[str], typing.Any]] = {
    bool: to_bool,
    bytes: lambda d: bytes(d, "utf-8"),
    bytearray: lambda d: bytearray(d, "utf-8"),
    datetime.datetime: to_datetime,
    hikari.Snowflake: parse_snowflake,
    urlparse.DefragResult: defragment_url,
    urlparse.ParseResult: parse_url,
    urlparse.SplitResult: split_url,
}


def override_type(cls: parsing.ConverterSig[typing.Any], /) -> parsing.ConverterSig[typing.Any]:
    return _TYPE_OVERRIDES.get(cls, cls)


to_channel: typing.Final[ToChannel] = ToChannel()
"""Convert user input to a `hikari.PartialChannel` object."""

to_colour: typing.Final[collections.Callable[[_ArgumentT], hikari.Color]] = to_color
"""Convert user input to a `hikari.Color` object."""

to_emoji: typing.Final[ToEmoji] = ToEmoji()
"""Convert user input to a cached `hikari.KnownCustomEmoji` object.

.. note::
    If you just want to convert inpute to a `hikari.Emoji`, `hikari.CustomEmoji`
    or `hikari.UnicodeEmoji` without making any cache or REST calls then you
    can just use the relevant `hikari.Emoji.parse`, `hikari.CustomEmoji.parse`
    or `hikari.UnicodeEmoji.parse` methods.
"""

to_guild: typing.Final[ToGuild] = ToGuild()
"""Convert user input to a `hikari.Guild` object."""

to_invite: typing.Final[ToInvite] = ToInvite()
"""Convert user input to a cached `hikari.InviteWithMetadata` object."""

to_invite_with_metadata: typing.Final[ToInviteWithMetadata] = ToInviteWithMetadata()
"""Convert user input to a `hikari.Invite` object."""

to_member: typing.Final[ToMember] = ToMember()
"""Convert user input to a `hikari.Member` object."""

to_presence: typing.Final[ToPresence] = ToPresence()
"""Convert user input to a cached `hikari.MemberPresence`."""

to_role: typing.Final[ToRole] = ToRole()
"""Convert user input to a `hikari.Role` object."""

to_snowflake: typing.Final[collections.Callable[[_ArgumentT], hikari.Snowflake]] = parse_snowflake
"""Convert user input to a `hikari.Snowflake`.

.. note::
    This also range validates the input.
"""

to_user: typing.Final[ToUser] = ToUser()
"""Convert user input to a `hikari.User` object."""

to_message: typing.Final[ToMessage] = ToMessage()
"""Convert user input to a `hikari.Message` object."""

to_voice_state: typing.Final[ToVoiceState] = ToVoiceState()
"""Convert user input to a cached `hikari.VoiceState`."""



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Collection of utility functions used within Tanjun."""
from __future__ import annotations

__all__: list[str] = [
    "ALL_PERMISSIONS",
    "CastedView",
    "DM_PERMISSIONS",
    "calculate_everyone_permissions",
    "calculate_permissions",
    "fetch_everyone_permissions",
    "fetch_permissions",
    "gather_checks",
    "match_prefix_names",
]

import asyncio
import typing
from collections import abc as collections

import hikari

from . import errors
from .dependencies import async_cache

if typing.TYPE_CHECKING:
    from . import abc

_KeyT = typing.TypeVar("_KeyT")
_OtherValueT = typing.TypeVar("_OtherValueT")
_ValueT = typing.TypeVar("_ValueT")


async def _execute_check(ctx: abc.Context, callback: abc.CheckSig, /) -> bool:
    foo = ctx.call_with_async_di(callback, ctx)
    if result := await foo:
        return result

    raise errors.FailedCheck


async def gather_checks(ctx: abc.Context, checks_: collections.Iterable[abc.CheckSig], /) -> bool:
    """Gather a collection of checks.

    Parameters
    ----------
    ctx : tanjun.abc.Context
        The context to check.
    checks : collections.abc.Iterable[tanjun.abc.CheckSig]
        An iterable of injectable checks.

    Returns
    -------
    bool
        Whether all the checks passed or not.
    """
    try:
        await asyncio.gather(*(_execute_check(ctx, check) for check in checks_))
        # InjectableCheck will raise FailedCheck if a false is received so if
        # we get this far then it's True.
        return True

    except errors.FailedCheck:
        return False


def match_prefix_names(content: str, names: collections.Iterable[str], /) -> typing.Optional[str]:
    """Search for a matching name in a string.

    Parameters
    ----------
    content : str
        The string to match names against.
    names : collections.abc.Iterable[str]
        The names to search for.

    Returns
    -------
    str | None
        The name that matched or None if no name matched.
    """
    for name in names:
        # Here we enforce that a name must either be at the end of content or be followed by a space. This helps
        # avoid issues with ambiguous naming where a command with the names "name" and "names" may sometimes hit
        # the former before the latter when triggered with the latter, leading to the command potentially being
        # inconsistently parsed.
        if content == name or content.startswith(name) and content[len(name)] == " ":
            return name


ALL_PERMISSIONS: typing.Final[hikari.Permissions] = hikari.Permissions.all_permissions()
"""All of all the known permissions based on the linked version of Hikari."""

DM_PERMISSIONS: typing.Final[hikari.Permissions] = (
    hikari.Permissions.ADD_REACTIONS
    | hikari.Permissions.VIEW_CHANNEL
    | hikari.Permissions.SEND_MESSAGES
    | hikari.Permissions.EMBED_LINKS
    | hikari.Permissions.ATTACH_FILES
    | hikari.Permissions.READ_MESSAGE_HISTORY
    | hikari.Permissions.USE_EXTERNAL_EMOJIS
    | hikari.Permissions.USE_EXTERNAL_STICKERS
    | hikari.Permissions.USE_APPLICATION_COMMANDS
)
"""Bitfield of the permissions which are accessibly within DM channels."""


def _calculate_channel_overwrites(
    channel: hikari.GuildChannel, member: hikari.Member, permissions: hikari.Permissions
) -> hikari.Permissions:
    if everyone_overwrite := channel.permission_overwrites.get(member.guild_id):
        permissions &= ~everyone_overwrite.deny
        permissions |= everyone_overwrite.allow

    deny = hikari.Permissions.NONE
    allow = hikari.Permissions.NONE

    for overwrite in filter(None, map(channel.permission_overwrites.get, member.role_ids)):
        deny |= overwrite.deny
        allow |= overwrite.allow

    permissions &= ~deny
    permissions |= allow

    if member_overwrite := channel.permission_overwrites.get(member.user.id):
        permissions &= ~member_overwrite.deny
        permissions |= member_overwrite.allow

    return permissions


def _calculate_role_permissions(
    roles: collections.Mapping[hikari.Snowflake, hikari.Role], member: hikari.Member
) -> hikari.Permissions:
    permissions = roles[member.guild_id].permissions

    for role in map(roles.get, member.role_ids):
        if role and role.id != member.guild_id:
            permissions |= role.permissions

    return permissions


# TODO: implicitly handle more special cases?
def calculate_permissions(
    member: hikari.Member,
    guild: hikari.Guild,
    roles: collections.Mapping[hikari.Snowflake, hikari.Role],
    *,
    channel: typing.Optional[hikari.GuildChannel] = None,
) -> hikari.Permissions:
    """Calculate the permissions a member has within a guild.

    Parameters
    ----------
    member : hikari.Member
        Object of the member to calculate the permissions for.
    guild : hikari.Guild
        Object of the guild to calculate their permissions within.
    roles : collections.abc.Mapping[hikari.Snowflake, hikari.Role]
        Mapping of snowflake IDs to objects of the roles within the target
        guild.

    Other Parameters
    ----------------
    channel : hikari.GuildChannel | None
        Object of the channel to calculate the member's permissions in.

        If this is left as `None` then this will just calculate their
        permissions on a guild level.

    Returns
    -------
    hikari.Permission
        Value of the member's permissions either within the guild or specified
        guild channel.
    """
    if member.guild_id != guild.id:
        raise ValueError("Member object isn't from the provided guild")

    # Guild owners are implicitly admins.
    if guild.owner_id == member.user.id:
        return ALL_PERMISSIONS

    # Admin permission overrides all overwrites and is only applicable to roles.
    if (permissions := _calculate_role_permissions(roles, member)) & permissions.ADMINISTRATOR:
        return ALL_PERMISSIONS

    if not channel:
        return permissions

    return _calculate_channel_overwrites(channel, member, permissions)


async def _fetch_channel(
    client: abc.Client, channel: hikari.SnowflakeishOr[hikari.PartialChannel]
) -> hikari.GuildChannel:
    if isinstance(channel, hikari.GuildChannel):
        return channel

    channel_id = hikari.Snowflake(channel)
    if client.cache and (found_channel_ := client.cache.get_guild_channel(channel_id)):
        return found_channel_

    if channel_cache := client.get_type_dependency(_ChannelCacheT):
        try:
            return await channel_cache.get(channel_id)

        except async_cache.EntryNotFound:
            raise

        except async_cache.CacheMissError:
            pass

    found_channel = await client.rest.fetch_channel(channel_id)
    assert isinstance(found_channel, hikari.GuildChannel), "Cannot perform operation on a DM channel."
    return found_channel


_ChannelCacheT = async_cache.SfCache[hikari.GuildChannel]
_GuildCacheT = async_cache.SfCache[hikari.Guild]
_RoleCacheT = async_cache.SfCache[hikari.Role]
_GuldRoleCacheT = async_cache.SfGuildBound[hikari.Role]


async def fetch_permissions(
    client: abc.Client,
    member: hikari.Member,
    /,
    *,
    channel: typing.Optional[hikari.SnowflakeishOr[hikari.PartialChannel]] = None,
) -> hikari.Permissions:
    """Calculate the permissions a member has within a guild.

    .. note::
        This callback will fallback to REST requests if cache lookups fail or
        are not possible.

    Parameters
    ----------
    client : tanjun.abc.Client
        The Tanjun client to use for lookups.
    member : hikari.Member
        The object of the member to calculate the permissions for.

    Other Parameters
    ----------------
    channel : hikari.SnowflakeishOr | hikari.GuildChannel | None
        The object of ID of the channel to get their permissions in.
        If left as `None` then this will return their base guild
        permissions.

    Returns
    -------
    hikari.Permissions
        The calculated permissions.
    """
    # The ordering of how this adds and removes permissions does matter.
    # For more information see https://discord.com/developers/docs/topics/permissions#permission-hierarchy.
    guild: typing.Optional[hikari.Guild]
    roles: typing.Optional[collections.Mapping[hikari.Snowflake, hikari.Role]] = None
    guild = client.cache.get_guild(member.guild_id) if client.cache else None
    if not guild and (guild_cache := client.get_type_dependency(_GuildCacheT)):
        try:
            guild = await guild_cache.get(member.guild_id)

        except async_cache.EntryNotFound:
            raise

        except async_cache.CacheMissError:
            pass

    if not guild:
        guild = await client.rest.fetch_guild(member.guild_id)
        roles = guild.roles

    # Guild owners are implicitly admins.
    if guild.owner_id == member.user.id:
        return ALL_PERMISSIONS

    roles = roles or client.cache and client.cache.get_roles_view_for_guild(member.guild_id)
    if not roles and (role_cache := client.get_type_dependency(_GuldRoleCacheT)):
        roles = {role.id: role for role in await role_cache.iter_for_guild(member.guild_id)}

    if not roles:
        raw_roles = await client.rest.fetch_roles(member.guild_id)
        roles = {role.id: role for role in raw_roles}

    # Admin permission overrides all overwrites and is only applicable to roles.
    if (permissions := _calculate_role_permissions(roles, member)) & permissions.ADMINISTRATOR:
        return ALL_PERMISSIONS

    if not channel:
        return permissions

    channel = await _fetch_channel(client, channel)
    if channel.guild_id != guild.id:
        raise ValueError("Channel doesn't match up with the member's guild")

    return _calculate_channel_overwrites(channel, member, permissions)


def calculate_everyone_permissions(
    everyone_role: hikari.Role,
    /,
    *,
    channel: typing.Optional[hikari.GuildChannel] = None,
) -> hikari.Permissions:
    """Calculate a guild's default permissions within the guild or for a specific channel.

    Parameters
    ----------
    everyone_role : hikari.Role
        The guild's default @everyone role.

    Other Parameters
    ----------------
    channel : hikari.GuildChannel | None
        The channel to calculate the permissions for.

        If this is left as `None` then this will just calculate the default
        permissions on a guild level.

    Returns
    -------
    hikari.Permissions
        The calculated permissions.
    """
    # The ordering of how this adds and removes permissions does matter.
    # For more information see https://discord.com/developers/docs/topics/permissions#permission-hierarchy.
    permissions = everyone_role.permissions
    # Admin permission overrides all overwrites and is only applicable to roles.
    if permissions & permissions.ADMINISTRATOR:
        return ALL_PERMISSIONS

    if not channel:
        return permissions

    if everyone_overwrite := channel.permission_overwrites.get(everyone_role.guild_id):
        permissions &= ~everyone_overwrite.deny
        permissions |= everyone_overwrite.allow

    return permissions


async def fetch_everyone_permissions(
    client: abc.Client,
    guild_id: hikari.Snowflake,
    /,
    *,
    channel: typing.Optional[hikari.SnowflakeishOr[hikari.PartialChannel]] = None,
) -> hikari.Permissions:
    """Calculate the permissions a guild's default @everyone role has within a guild or for a specific channel.

    .. note::
        This callback will fallback to REST requests if cache lookups fail or
        are not possible.

    Parameters
    ----------
    client : tanjun.abc.Client
        The Tanjun client to use for lookups.
    guild_id : hikari.Snowflake
        ID of the guild to calculate the default permissions for.

    Other Parameters
    ----------------
    channel : hikari.SnowflakeishOr | hikari.PartialChannel | None
        The channel to calculate the permissions for.

        If this is left as `None` then this will just calculate the default
        permissions on a guild level.


    Returns
    -------
    hikari.Permissions
        The calculated permissions.
    """
    # The ordering of how this adds and removes permissions does matter.
    # For more information see https://discord.com/developers/docs/topics/permissions#permission-hierarchy.
    role = client.cache.get_role(guild_id) if client.cache else None
    if not role and (role_cache := client.get_type_dependency(_RoleCacheT)):
        try:
            role = await role_cache.get(guild_id)

        except async_cache.EntryNotFound:
            raise

        except async_cache.CacheMissError:
            pass

    if not role:
        for role in await client.rest.fetch_roles(guild_id):
            if role.id == guild_id:
                break

        else:
            raise RuntimeError("Failed to find guild's @everyone role")

    permissions = role.permissions
    # Admin permission overrides all overwrites and is only applicable to roles.
    if permissions & permissions.ADMINISTRATOR:
        return ALL_PERMISSIONS

    if not channel:
        return permissions

    channel = await _fetch_channel(client, channel)
    if everyone_overwrite := channel.permission_overwrites.get(guild_id):
        permissions &= ~everyone_overwrite.deny
        permissions |= everyone_overwrite.allow

    return permissions


class CastedView(collections.Mapping[_KeyT, _OtherValueT], typing.Generic[_KeyT, _ValueT, _OtherValueT]):
    __slots__ = ("_buffer", "_cast", "_raw_data")

    def __init__(self, raw_data: dict[_KeyT, _ValueT], cast: collections.Callable[[_ValueT], _OtherValueT]) -> None:
        self._buffer: dict[_KeyT, _OtherValueT] = {}
        self._cast = cast
        self._raw_data = raw_data

    def __getitem__(self, key: _KeyT, /) -> _OtherValueT:
        try:
            return self._buffer[key]

        except KeyError:
            pass

        entry = self._raw_data[key]
        result = self._cast(entry)
        self._buffer[key] = result
        return result

    def __iter__(self) -> collections.Iterator[_KeyT]:
        return iter(self._raw_data)

    def __len__(self) -> int:
        return len(self._raw_data)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""Example component which takes advantage of dependency injection."""
import datetime
import typing

import tanjun
from examples import protos

component = tanjun.Component()


@component.with_command
@tanjun.with_guild_check
@tanjun.as_message_command("guild")
async def guild_command(
    # Here we ask for an implementation of the type `DatabaseProto` to be
    # provided as an argument whenever the client calls this command.
    #
    # The implementation for this is provided with Client.set_type_dependency.
    ctx: tanjun.abc.MessageContext,
    database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto),
):
    assert ctx.guild_id is not None  # This is checked by the "with_guild_check"
    guild_info = await database.get_guild_info(ctx.guild_id)

    if not guild_info:
        # HaltExecution will lead to the command being marked as not found,
        # essentially hiding it.
        raise tanjun.HaltExecution

    ...  # TODO: implement response


@component.with_command
@tanjun.as_message_command_group("user")
async def user(
    # Here we ask for an implementation of the type `DatabaseProto` to be
    # provided as an argument whenever the client calls this command.
    #
    # The implementation for this is provided with Client.set_type_dependency.
    ctx: tanjun.abc.MessageContext,
    database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto),
) -> None:
    user = await database.get_user_info(ctx.author.id)

    if not user:
        # CommandError's message will be sent as a response message.
        raise tanjun.CommandError("No information stored for you")

    ...  # TODO: implement response


@user.with_command
@tanjun.as_message_command("remove self")
async def remove_self(
    # Here we ask for an implementation of the type `DatabaseProto` to be
    # provided as an argument whenever the client calls this command.
    #
    # The implementation for this is provided with Client.set_type_dependency.
    ctx: tanjun.abc.MessageContext,
    database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto),
) -> None:
    await database.remove_user(ctx.author.id)


# Since this is being used as an injected callback, we can also ask for an injected type here.
async def _fetch_info(database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto)) -> typing.Any:
    raise NotImplementedError  # This is an example callback and doesn't provide an implementation.


# Since this is being used as an injected callback, we can also ask for an injected type here.
async def _fetch_cachable_info(
    database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto),
) -> typing.Any:
    raise NotImplementedError  # This is an example callback and doesn't provide an implementation.


@component.with_command
@tanjun.as_slash_command("info", "Command description")
async def get_info(
    ctx: tanjun.abc.SlashContext,
    # Here we set _fetch_info as an injected callback.
    #
    # Injected callbacks are callbacks which'll be called before this function is called
    # with the result of it being being passed to the command callback.
    # (note these also support dependency injection).
    info: typing.Any = tanjun.inject(callback=_fetch_info),
    # Here we set _fetch_cachable_info as a cached injected callback.
    #
    # `cached_inject(callback)` is a variant of `inject(callback=callback)`
    # which caches the result of the callback for the provided expire duration.
    cached_info: typing.Any = tanjun.cached_inject(_fetch_cachable_info, expire_after=datetime.timedelta(minutes=30)),
) -> None:
    await ctx.respond(cached_info.format())
    await ctx.respond(info.format())


# Here we define a loader which can be used to easily load this example
# components into a bot from a link (assuming the environment has all the
# right configurations setup.)
@tanjun.as_loader
def load_examples(client: tanjun.Client) -> None:
    client.add_component(component.copy())


# Here we define an unloader which can be used to easily unload and reload
# this example components in a bot from a link.
@tanjun.as_unloader
def unload_examples(client: tanjun.Client) -> None:
    # Since there's no guarantee the stored component will still be the
    # same as component, we remove it by name.
    client.remove_component_by_name(component.name)



# -*- coding: utf-8 -*-
# cython: language_level=3
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""Example of how to run a standard Tanjun client instance with a GatewayBot."""
from collections import abc as collections

import hikari

import tanjun
from examples import config
from examples import impls
from examples import protos


async def get_prefix(
    ctx: tanjun.abc.MessageContext, database: protos.DatabaseProto = tanjun.inject(type=protos.DatabaseProto)
) -> collections.Sequence[str]:
    if ctx.guild_id and (guild_info := await database.get_guild_info(ctx.guild_id)):
        return guild_info.prefixes

    return ()


def run() -> None:
    loaded_config = config.ExampleConfig.load()
    # Note that by default `from_gateway_bot` sets `event_managed` to `True`.
    # This means that the client will be implicitly started and stopped
    # based on Hikari's lifetime events.
    #
    # You can alternatively start and stop the Tanjun client yourself
    # by calling `open` and `close` on it or using it as a context manager.
    #
    # Note that starting a Tanjun client before the relevant bot instance
    # may lead to erroneous behaviour as it won't be able to make requests.
    bot = hikari.GatewayBot(loaded_config.bot_token)
    database = impls.DatabaseImpl()
    (
        tanjun.Client.from_gateway_bot(bot)
        .load_modules("examples.complex_component")
        # Both slash commands and message commands can be automatically executed
        # by a gateway bot bound client
        .load_modules("examples.message_component")
        .load_modules("examples.slash_component")
        .add_prefix(loaded_config.prefix)
        .set_prefix_getter(get_prefix)
        .set_type_dependency(config.ExampleConfig, loaded_config)
        .set_type_dependency(protos.DatabaseProto, database)
        # Here we use client callbacks to manage the database, STOPPING can also be used to stop it.
        .add_client_callback(tanjun.ClientCallbackNames.STARTING, database.connect)
    )
    bot.run()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# cython: language_level=3
# Tanjun Examples - A collection of examples for Tanjun.
# Written in 2021 by Lucina Lucina@lmbyrne.dev
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain worldwide.
# This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along with this software.
# If not, see <https://creativecommons.org/publicdomain/zero/1.0/>.
"""Example basic usage of Tanjun's slash command system."""

# As a note, a slash command context should always be responded to as leaving
# it without a response will leave the command call either marked as loading
# for 15 minutes (if it's been deferred) or marked as failed after
# 3 seconds (by default slash commands are automatically deferred before the
# 3 second expire mark but this doesn't negate the need to give an actual
# response).
#
# For most of these examples `tanjun.abc.SlashContext.respond` is used for
# responding to interactions, this abstracts away the interactions response
# flow by tracking state on the Context object in-order to decide how to
# respond. While the lower level approache may also be used for responding it
# is generally recommended that you use the relevant lower level methods on
# `tanjun.abc.SlashContext` to allow state to still be tracked and ensure
# better compatibility.
import asyncio
import typing

import hikari

import tanjun

component = tanjun.Component()


# Here we declare a command as defaulting to ephemeral, this means that all
# calls made to SlashContext's message creating methods (e.g. respond,
# create_initial_respones and create_followup) will be ephemeral (only visible
# to the user who called the command) unless `flags` is specified. This includes
# calls made by command checks and by the CommandError handler.
@component.with_slash_command
@tanjun.as_slash_command("nsfw", "command description", default_to_ephemeral=True)
async def nsfw_command(ctx: tanjun.abc.Context) -> None:
    # Thanks to default_to_ephemeral=True, this response will be ephemeral.
    await ctx.respond("nsfw stuff")


# While slash command groups may be nested, this only works up to one level.
top_group = component.with_slash_command(tanjun.slash_command_group("places", "get info about places"))
nested_group = top_group.with_command(tanjun.slash_command_group("interaction", "say hello or something!"))


@nested_group.with_command
# This adds a required member option to the command which'll be passed to the
# "member" argument as type hikari.Member.
@tanjun.with_member_slash_option("member", "Option description")
# This adds an optional string option to the command which'll be passed
# to the "name" argument as type str if it was provided else None.
@tanjun.with_str_slash_option("name", "Option description", default=None)
@tanjun.as_slash_command("hi", "command description")
async def hi_command(ctx: tanjun.abc.Context, name: typing.Optional[str], member: hikari.Member) -> None:
    if name:
        await ctx.respond(f"Hi, {name} and {member.username}")

    else:
        await ctx.respond(f"Hi {member.username}")


@top_group.with_command
# Here we add a required string option which'll be convertered to an emoji object.
@tanjun.with_str_slash_option("emoji", "Option description", converters=tanjun.to_emoji)
@tanjun.as_slash_command("japan", "command description")
async def japan_command(ctx: tanjun.abc.Context, emoji: hikari.Emoji) -> None:
    await ctx.respond(f"Nihongo ga dekimasu ka? {emoji}")


@top_group.with_command
@tanjun.as_slash_command("europe", "command description")
async def europe_command(ctx: tanjun.abc.Context) -> None:
    await ctx.respond("I don't know how to describe Europe... small?")


# The previous code leads to top_group having the following structure
#
# places (command group)
# |
# |__ japan (command)
# |
# |__ europe (command)
# |
# |__ interaction (command group)
#    |
#    |__ hi (command)


# For more information on slash command options and converters, see the
# documentation for `tanjun.commands` and `tanjun.converters`.
#
# It's worth noting that converters are only supported for string, integer
# and float options.


@component.with_command
@tanjun.as_slash_command("lower", "Lower level command which takes advantage of slash command specific detail")
async def lower_command(ctx: tanjun.abc.SlashContext) -> None:
    # Since `SlashContext.respond` can't have `flags` as an argument, providing
    # the flags when creating the initial response requires lower level usage.
    #
    # As a note, you can only create the initial response for a slash command
    # context once and any further calls will result in an error being raised.
    # To create follow up responses see `SlashContext.create_followup`.
    #
    # As another note, an initial response for a slash context must be created
    # within 3 seconds of the interaction being received otherwise it will
    # either be automatically deferred or expire (if automatic deferral is
    # disabled). In the case that an interaction is deferred then
    # `SlashContext.edit_initial_response` or `SlashContext.respond` should be
    # used to edit an initial response in. `tanjun.abc.SlashContext.defer` may
    # be used to defer an interaction in the case that automatic deferral is
    # disabled or the response is always going to be deferred.
    await ctx.create_initial_response("I'm sorry, Dave", flags=hikari.MessageFlag.EPHEMERAL, tts=True)

    # Since `SlashContext.respond` can't have attachments as an argument,
    # providing `attachments` requires the usage of slash command specific and
    # lower level detail.
    #
    # As a note, you can only create followup responses after an initial response
    # has been made and any pre-mature calls will result in errors being raised.
    await ctx.create_followup(
        "I'm afraid I can't do that",
        attachments=[hikari.URL("https://cdn.discordapp.com/emojis/520742867933724676.png?v=1")],
    )


# Here's a lower level usage example which explicitly defers the initial response.
# This will have to be done within clients where automatic deferral is disabled
# if a command (including the relevant converters and checks) takes more than 3
# seconds to execute to avoid errenous behaviour.
@component.with_command
@tanjun.as_slash_command("defer", "Lower level command which explicitly defers")
async def defer_command(ctx: tanjun.abc.SlashContext) -> None:
    # Note that if we want the response that's later edited in to be ephemeral
    # then we can pass `flags=hikari.MessageFlags.EPHEMERAL` to `SlashContext.defer`.
    await ctx.defer()
    await asyncio.sleep(5)  # Do some work which may take a while
    # Either edit_initial_response or respond may be used here.
    await ctx.edit_initial_response("Done 👍")


# Here we use make_loader to define a loader which can be used to both load
# and unload and reload this example component into a bot from a link
# (assuming the environment has all the right configurations setup).
#
# Alternatively @tanjun.as_loader and @tanjun.as_unloader can be used
# for more fine-grained control.
load_slash = component.make_loader()



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""A flexible command framework designed to extend Hikari.

Examples
--------
A Tanjun client can be quickly initialised from a Hikari gateway bot through
`tanjun.Client.from_gateway_bot`, this enables both slash (interaction) and message
command execution:

```py
bot = hikari.GatewayBot("BOT_TOKEN")

# As a note, unless event_managed=False is passed here then this client
# will be managed based on gateway startup and stopping events.
# mention_prefix=True instructs the client to also set mention prefixes on the
# first startup.
client = tanjun.Client.from_gateway_bot(bot, declare_global_commands=True, mention_prefix=True)

component = tanjun.Component()
client.add_component(component)

# Declare a message command with some basic parser logic.
@component.with_command
@tanjun.with_greedy_argument("name", default="World")
@tanjun.as_message_command("test")
async def test_command(ctx: tanjun.abc.Context, name: str) -> None:
    await ctx.respond(f"Hello, {name}!")

# Declare a ping slash command
@component.with_command
@tanjun.with_user_slash_option("user", "The user facing command option's description", default=None)
@tanjun.as_slash_command("hello", "The command's user facing description")
async def hello(ctx: tanjun.abc.Context, user: hikari.User | None) -> None:
    user = user or ctx.author
    await ctx.respond(f"Hello, {user}!")
```

Alternatively, the client can also be built from a RESTBot but this will only
enable slash (interaction) command execution:

```py
bot = hikari.RESTBot("BOT_TOKEN", "Bot")

# declare_global_commands=True instructs the client to set the global commands
# for the relevant bot on first startup (this will replace any previously
# declared commands).
client = tanjun.Client.from_rest_bot(bot, declare_global_commands=True)

# This will load components from modules based on loader functions.
# For more information on this see `tanjun.as_loader`.
client.load_modules("module.paths")

# Note, unlike a gateway bound bot, the rest bot will not automatically start
# itself due to the lack of Hikari lifetime events in this environment and
# will have to be started after the Hikari client.
async def main() -> None:
    await bot.start()
    async with client.open():
        await bot.join()
```

For more extensive examples see the
[repository's examples](https://github.com/FasterSpeeding/Tanjun/tree/master/examples).

There are also
[written tutorials](https://patchwork.systems/programming/hikari-discord-bot/index.html)
that cover making a bot from scratch through to advanced concepts like Dependency Injection.
"""
from __future__ import annotations

__all__: list[str] = [
    "AnyHooks",
    "BucketResource",
    "Client",
    "ClientCallbackNames",
    "CommandError",
    "Component",
    "ConversionError",
    "FailedCheck",
    "FailedModuleLoad",
    "FailedModuleUnload",
    "HaltExecution",
    "Hooks",
    "InMemoryConcurrencyLimiter",
    "InMemoryCooldownManager",
    "LazyConstant",
    "MenuCommand",
    "MessageAcceptsEnum",
    "MessageCommand",
    "MessageCommandGroup",
    "MessageHooks",
    "MissingDependencyError",
    "ModuleMissingLoaders",
    "ModuleStateConflict",
    "NotEnoughArgumentsError",
    "ParserError",
    "ShlexParser",
    "SlashCommand",
    "SlashCommandGroup",
    "SlashHooks",
    "TanjunError",
    "TooManyArgumentsError",
    "__author__",
    "__ci__",
    "__copyright__",
    "__coverage__",
    "__docs__",
    "__email__",
    "__issue_tracker__",
    "__license__",
    "__url__",
    "__version__",
    "abc",
    "as_interval",
    "as_loader",
    "as_message_command",
    "as_message_command_group",
    "as_message_menu",
    "as_self_injecting",
    "as_slash_command",
    "as_unloader",
    "as_user_menu",
    "cached_inject",
    "checks",
    "clients",
    "commands",
    "components",
    "context",
    "conversion",
    "dependencies",
    "errors",
    "hooks",
    "inject",
    "inject_lc",
    "injected",
    "injecting",
    "parsing",
    "schedules",
    "slash_command_group",
    "to_bool",
    "to_channel",
    "to_color",
    "to_colour",
    "to_datetime",
    "to_emoji",
    "to_guild",
    "to_invite",
    "to_invite_with_metadata",
    "to_member",
    "to_message",
    "to_presence",
    "to_role",
    "to_snowflake",
    "to_user",
    "to_voice_state",
    "utilities",
    "with_all_checks",
    "with_any_checks",
    "with_argument",
    "with_attachment_slash_option",
    "with_author_permission_check",
    "with_bool_slash_option",
    "with_channel_slash_option",
    "with_check",
    "with_concurrency_limit",
    "with_cooldown",
    "with_dm_check",
    "with_float_slash_option",
    "with_greedy_argument",
    "with_guild_check",
    "with_int_slash_option",
    "with_member_slash_option",
    "with_mentionable_slash_option",
    "with_multi_argument",
    "with_multi_option",
    "with_nsfw_check",
    "with_option",
    "with_own_permission_check",
    "with_owner_check",
    "with_parser",
    "with_role_slash_option",
    "with_sfw_check",
    "with_str_slash_option",
    "with_user_slash_option",
]

import typing

from alluka import inject
from alluka import inject as injected

from . import abc
from . import context
from . import utilities
from .abc import ClientCallbackNames
from .checks import with_all_checks
from .checks import with_any_checks
from .checks import with_author_permission_check
from .checks import with_check
from .checks import with_dm_check
from .checks import with_guild_check
from .checks import with_nsfw_check
from .checks import with_own_permission_check
from .checks import with_owner_check
from .checks import with_sfw_check
from .clients import Client
from .clients import MessageAcceptsEnum
from .clients import as_loader
from .clients import as_unloader
from .commands import MenuCommand
from .commands import MessageCommand
from .commands import MessageCommandGroup
from .commands import SlashCommand
from .commands import SlashCommandGroup
from .commands import as_message_command
from .commands import as_message_command_group
from .commands import as_message_menu
from .commands import as_slash_command
from .commands import as_user_menu
from .commands import slash_command_group
from .commands import with_attachment_slash_option
from .commands import with_bool_slash_option
from .commands import with_channel_slash_option
from .commands import with_float_slash_option
from .commands import with_int_slash_option
from .commands import with_member_slash_option
from .commands import with_mentionable_slash_option
from .commands import with_role_slash_option
from .commands import with_str_slash_option
from .commands import with_user_slash_option
from .components import Component
from .conversion import to_bool
from .conversion import to_channel
from .conversion import to_color
from .conversion import to_colour
from .conversion import to_datetime
from .conversion import to_emoji
from .conversion import to_guild
from .conversion import to_invite
from .conversion import to_invite_with_metadata
from .conversion import to_member
from .conversion import to_message
from .conversion import to_presence
from .conversion import to_role
from .conversion import to_snowflake
from .conversion import to_user
from .conversion import to_voice_state
from .dependencies import BucketResource
from .dependencies import InMemoryConcurrencyLimiter
from .dependencies import InMemoryCooldownManager
from .dependencies import LazyConstant
from .dependencies import cached_inject
from .dependencies import inject_lc
from .dependencies import with_concurrency_limit
from .dependencies import with_cooldown
from .errors import CommandError
from .errors import ConversionError
from .errors import FailedCheck
from .errors import FailedModuleLoad
from .errors import FailedModuleUnload
from .errors import HaltExecution
from .errors import MissingDependencyError
from .errors import ModuleMissingLoaders
from .errors import ModuleStateConflict
from .errors import NotEnoughArgumentsError
from .errors import ParserError
from .errors import TanjunError
from .errors import TooManyArgumentsError
from .hooks import AnyHooks
from .hooks import Hooks
from .hooks import MessageHooks
from .hooks import SlashHooks
from .injecting import as_self_injecting
from .parsing import ShlexParser
from .parsing import with_argument
from .parsing import with_greedy_argument
from .parsing import with_multi_argument
from .parsing import with_multi_option
from .parsing import with_option
from .parsing import with_parser
from .schedules import as_interval

__author__: typing.Final[str] = "Faster Speeding"
__ci__: typing.Final[str] = "https://github.com/FasterSpeeding/Tanjun/actions"
__copyright__: typing.Final[str] = "© 2020-2022 Faster Speeding"
__coverage__: typing.Final[str] = "https://codeclimate.com/github/FasterSpeeding/Tanjun"
__docs__: typing.Final[str] = "https://tanjun.cursed.solutions/"
__email__: typing.Final[str] = "lucina@lmbyrne.dev"
__issue_tracker__: typing.Final[str] = "https://github.com/FasterSpeeding/Tanjun/issues"
__license__: typing.Final[str] = "BSD"
__url__: typing.Final[str] = "https://github.com/FasterSpeeding/Tanjun"
__version__: typing.Final[str] = "2.4.3a1"



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# pyright: reportUnknownMemberType=none
# pyright: reportPrivateUsage=none
# This leads to too many false-positives around mocks.

import types
import typing
from unittest import mock

import alluka
import hikari
import pytest
from hikari import traits

import tanjun
from tanjun.context import base as base_context

_T = typing.TypeVar("_T")


def stub_class(cls: type[_T], /, **namespace: typing.Any) -> type[_T]:
    namespace["__slots__"] = ()

    for name in getattr(cls, "__abstractmethods__", None) or ():
        if name not in namespace:
            namespace[name] = mock.MagicMock()

    name = origin.__name__ if (origin := getattr(cls, "__origin__", None)) else cls.__name__
    new_cls = types.new_class(name, (cls,), exec_body=lambda body: body.update(namespace))
    return typing.cast(type[_T], new_cls)


@pytest.fixture()
def mock_client() -> tanjun.abc.Client:
    return mock.MagicMock(tanjun.abc.Client, rest=mock.AsyncMock(hikari.api.RESTClient))


@pytest.fixture()
def mock_component() -> tanjun.abc.Component:
    return mock.MagicMock(tanjun.abc.Component)


class TestBaseContext:
    @pytest.fixture()
    def context(
        self,
        mock_client: mock.Mock,
    ) -> base_context.BaseContext:
        return stub_class(base_context.BaseContext)(mock_client)

    def test_cache_property(self, context: tanjun.abc.Context, mock_client: mock.Mock):
        assert context.cache is mock_client.cache

    def test_client_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.client is mock_client

    def test_component_property(self, context: base_context.BaseContext, mock_component: tanjun.abc.Component):
        assert context.component is None

    def test_events_proprety(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.events is mock_client.events

    def test_rest_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.rest is mock_client.rest

    def test_server_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.server is mock_client.server

    def test_shard_property(self, mock_client: mock.Mock):
        mock_shard = mock.Mock()
        mock_client.shards = mock.MagicMock(spec=traits.ShardAware, shard_count=5, shards={2: mock_shard})
        context = stub_class(base_context.BaseContext, guild_id=hikari.Snowflake(123321123312))(mock_client)

        assert context.shard is mock_shard

    def test_shard_property_when_dm(self, mock_client: mock.Mock):
        mock_shard = mock.Mock()
        mock_client.shards = mock.Mock(shards={0: mock_shard})
        context = stub_class(base_context.BaseContext, guild_id=None)(mock_client)

        assert context.shard is mock_shard

    def test_shard_property_when_no_shards(self, context: tanjun.context.MessageContext):
        context._client = mock.Mock(shards=None)

        assert context.shard is None

    def test_shards_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.shards is mock_client.shards

    def test_voice_property(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert context.voice is mock_client.voice

    def test_finalise(self, context: base_context.BaseContext):
        context.finalise()
        assert context._final is True

    def test_set_component(self, context: base_context.BaseContext):
        component = mock.Mock()

        assert context.set_component(component) is context

        assert context.component is component
        assert context.get_type_dependency(tanjun.abc.Component) is component

    def test_set_component_when_none_and_previously_set(self, context: base_context.BaseContext):
        assert isinstance(context.injection_client.get_type_dependency, mock.Mock)
        context.injection_client.get_type_dependency.return_value = alluka.abc.UNDEFINED
        mock_component = mock.Mock()
        context.set_component(mock_component)
        context.set_component(None)

        assert context.component is None
        assert context.get_type_dependency(tanjun.abc.Component) is alluka.abc.UNDEFINED

    def test_set_component_when_none(self, context: base_context.BaseContext):
        assert isinstance(context.injection_client.get_type_dependency, mock.Mock)
        context.injection_client.get_type_dependency.return_value = alluka.abc.UNDEFINED
        context.set_component(None)
        context.set_component(None)

        assert context.component is None
        assert context.get_type_dependency(tanjun.abc.Component) is alluka.abc.UNDEFINED

    def test_set_component_when_final(self, context: base_context.BaseContext):
        component = mock.Mock()
        context.finalise()

        with pytest.raises(TypeError):
            context.set_component(component)

        assert context.component is not component

    def test_get_channel(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        mock_client.cache.get_guild_channel.return_value = mock.Mock(hikari.TextableGuildChannel)

        assert context.get_channel() is mock_client.cache.get_guild_channel.return_value

        mock_client.cache.get_guild_channel.assert_called_once_with(context.channel_id)

    def test_get_channel_when_cache_returns_none(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        mock_client.cache.get_guild_channel.return_value = None

        assert context.get_channel() is None

        mock_client.cache.get_guild_channel.assert_called_once_with(context.channel_id)

    def test_get_channel_when_cacheless(self):
        context = stub_class(base_context.BaseContext, guild_id=None)(mock.Mock(cache=None))

        assert context.get_channel() is None

    def test_get_guild(self, context: base_context.BaseContext, mock_client: mock.Mock):
        assert mock_client.cache is not None
        assert context.get_guild() is mock_client.cache.get_guild.return_value
        mock_client.cache.get_guild.assert_called_once_with(context.guild_id)

    def test_get_guild_when_cacheless(self):
        context = stub_class(base_context.BaseContext, guild_id=None)(mock.Mock(cache=None))

        assert context.get_guild() is None

    def test_get_guild_when_dm_bound(self):
        mock_client = mock.MagicMock()
        context = stub_class(base_context.BaseContext, guild_id=None)(mock_client)

        assert context.get_guild() is None
        mock_client.cache.get_guild.assert_not_called()

    @pytest.mark.asyncio()
    async def test_fetch_channel(self, context: base_context.BaseContext, mock_client: mock.Mock):
        mock_client.rest.fetch_channel.return_value = mock.Mock(hikari.TextableChannel)

        result = await context.fetch_channel()

        assert result is mock_client.rest.fetch_channel.return_value
        mock_client.rest.fetch_channel.assert_called_once_with(context.channel_id)

    @pytest.mark.asyncio()
    async def test_fetch_guild(self, context: base_context.BaseContext, mock_client: mock.Mock):
        result = await context.fetch_guild()

        assert result is mock_client.rest.fetch_guild.return_value
        mock_client.rest.fetch_guild.assert_called_once_with(context.guild_id)

    @pytest.mark.asyncio()
    async def test_fetch_guild_when_dm_bound(self, mock_client: mock.Mock):
        context = stub_class(base_context.BaseContext, guild_id=None)(mock_client)

        result = await context.fetch_guild()

        assert result is None
        mock_client.rest.fetch_guild.assert_not_called()



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020-2022, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Standard message command execution context implementations."""
from __future__ import annotations

__all__: list[str] = ["MessageContext"]

import asyncio
import datetime
import logging
import typing

import hikari

from .. import abc as tanjun_abc
from . import base

if typing.TYPE_CHECKING:
    from collections import abc as collections

    _MessageContextT = typing.TypeVar("_MessageContextT", bound="MessageContext")

_LOGGER = logging.getLogger("hikari.tanjun.context")


def _delete_after_to_float(delete_after: typing.Union[datetime.timedelta, float, int]) -> float:
    return delete_after.total_seconds() if isinstance(delete_after, datetime.timedelta) else float(delete_after)


class MessageContext(base.BaseContext, tanjun_abc.MessageContext):
    """Standard implementation of a command context as used within Tanjun."""

    __slots__ = (
        "_command",
        "_content",
        "_initial_response_id",
        "_last_response_id",
        "_response_lock",
        "_message",
        "_triggering_name",
        "_triggering_prefix",
    )

    def __init__(
        self,
        client: tanjun_abc.Client,
        content: str,
        message: hikari.Message,
        *,
        triggering_name: str = "",
        triggering_prefix: str = "",
    ) -> None:
        if message.content is None:
            raise ValueError("Cannot spawn context with a content-less message.")

        super().__init__(client)
        self._command: typing.Optional[tanjun_abc.MessageCommand[typing.Any]] = None
        self._content = content
        self._initial_response_id: typing.Optional[hikari.Snowflake] = None
        self._last_response_id: typing.Optional[hikari.Snowflake] = None
        self._response_lock = asyncio.Lock()
        self._message = message
        self._triggering_name = triggering_name
        self._triggering_prefix = triggering_prefix
        self._set_type_special_case(tanjun_abc.MessageContext, self)._set_type_special_case(MessageContext, self)

    def __repr__(self) -> str:
        return f"MessageContext <{self._message!r}, {self._command!r}>"

    @property
    def author(self) -> hikari.User:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._message.author

    @property
    def channel_id(self) -> hikari.Snowflake:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._message.channel_id

    @property
    def command(self) -> typing.Optional[tanjun_abc.MessageCommand[typing.Any]]:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        return self._command

    @property
    def content(self) -> str:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        return self._content

    @property
    def created_at(self) -> datetime.datetime:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._message.created_at

    @property
    def guild_id(self) -> typing.Optional[hikari.Snowflake]:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._message.guild_id

    @property
    def has_responded(self) -> bool:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._initial_response_id is not None

    @property
    def is_human(self) -> bool:
        # <<inherited docstring from tanjun.abc.Context>>.
        return not self._message.author.is_bot and self._message.webhook_id is None

    @property
    def member(self) -> typing.Optional[hikari.Member]:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._message.member

    @property
    def message(self) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        return self._message

    @property
    def triggering_name(self) -> str:
        # <<inherited docstring from tanjun.abc.Context>>.
        return self._triggering_name

    @property
    def triggering_prefix(self) -> str:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        return self._triggering_prefix

    def set_command(
        self: _MessageContextT, command: typing.Optional[tanjun_abc.MessageCommand[typing.Any]], /
    ) -> _MessageContextT:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        self._assert_not_final()
        if command:
            # TODO: command group?
            self._set_type_special_case(tanjun_abc.ExecutableCommand, command)._set_type_special_case(
                tanjun_abc.MessageCommand, command
            )

        elif self._command:
            self._remove_type_special_case(tanjun_abc.ExecutableCommand)._remove_type_special_case(
                tanjun_abc.MessageCommand
            )

        self._command = command
        return self

    def set_content(self: _MessageContextT, content: str, /) -> _MessageContextT:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        self._assert_not_final()
        self._content = content
        return self

    def set_triggering_name(self: _MessageContextT, name: str, /) -> _MessageContextT:
        # <<inherited docstring from tanjun.abc.MessageContext>>.
        self._assert_not_final()
        self._triggering_name = name
        return self

    def set_triggering_prefix(self: _MessageContextT, triggering_prefix: str, /) -> _MessageContextT:
        """Set the triggering prefix for this context.

        Parameters
        ----------
        triggering_prefix : str
            The triggering prefix to set.

        Returns
        -------
        Self
            This context to allow for chaining.
        """
        self._assert_not_final()
        self._triggering_prefix = triggering_prefix
        return self

    async def delete_initial_response(self) -> None:
        # <<inherited docstring from tanjun.abc.Context>>.
        if self._initial_response_id is None:
            raise LookupError("Context has no initial response")

        await self._client.rest.delete_message(self._message.channel_id, self._initial_response_id)

    async def delete_last_response(self) -> None:
        # <<inherited docstring from tanjun.abc.Context>>.
        if self._last_response_id is None:
            raise LookupError("Context has no previous responses")

        await self._client.rest.delete_message(self._message.channel_id, self._last_response_id)

    async def edit_initial_response(
        self,
        content: hikari.UndefinedOr[typing.Any] = hikari.UNDEFINED,
        *,
        delete_after: typing.Union[datetime.timedelta, float, int, None] = None,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[collections.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedNoneOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedNoneOr[collections.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedNoneOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedNoneOr[collections.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        replace_attachments: bool = False,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialUser], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
        role_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialRole], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
    ) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.Context>>.
        delete_after = _delete_after_to_float(delete_after) if delete_after is not None else None
        if self._initial_response_id is None:
            raise LookupError("Context has no initial response")

        message = await self.rest.edit_message(
            self._message.channel_id,
            self._initial_response_id,
            content=content,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            embed=embed,
            embeds=embeds,
            replace_attachments=replace_attachments,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )
        if delete_after is not None:
            asyncio.create_task(self._delete_after(delete_after, message))

        return message

    async def edit_last_response(
        self,
        content: hikari.UndefinedOr[typing.Any] = hikari.UNDEFINED,
        *,
        delete_after: typing.Union[datetime.timedelta, float, int, None] = None,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[collections.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedNoneOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedNoneOr[collections.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedNoneOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedNoneOr[collections.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        replace_attachments: bool = False,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialUser], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
        role_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialRole], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
    ) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.Context>>.
        delete_after = _delete_after_to_float(delete_after) if delete_after is not None else None
        if self._last_response_id is None:
            raise LookupError("Context has no previous tracked response")

        message = await self.rest.edit_message(
            self._message.channel_id,
            self._last_response_id,
            content=content,
            attachment=attachment,
            attachments=attachments,
            component=component,
            components=components,
            embed=embed,
            embeds=embeds,
            replace_attachments=replace_attachments,
            mentions_everyone=mentions_everyone,
            user_mentions=user_mentions,
            role_mentions=role_mentions,
        )

        if delete_after is not None:
            asyncio.create_task(self._delete_after(delete_after, message))

        return message

    async def fetch_initial_response(self) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.Context>>.
        if self._initial_response_id is not None:
            return await self.client.rest.fetch_message(self._message.channel_id, self._initial_response_id)

        raise LookupError("No initial response found for this context")

    async def fetch_last_response(self) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.Context>>.
        if self._last_response_id is not None:
            return await self.client.rest.fetch_message(self._message.channel_id, self._last_response_id)

        raise LookupError("No responses found for this context")

    @staticmethod
    async def _delete_after(delete_after: float, message: hikari.Message) -> None:
        await asyncio.sleep(delete_after)
        try:
            await message.delete()
        except hikari.NotFoundError as exc:
            _LOGGER.debug("Failed to delete response message after %.2f seconds", delete_after, exc_info=exc)

    async def respond(
        self,
        content: hikari.UndefinedOr[typing.Any] = hikari.UNDEFINED,
        *,
        ensure_result: bool = True,
        delete_after: typing.Union[datetime.timedelta, float, int, None] = None,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[collections.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[collections.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedOr[collections.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        tts: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        nonce: hikari.UndefinedOr[str] = hikari.UNDEFINED,
        reply: typing.Union[bool, hikari.SnowflakeishOr[hikari.PartialMessage], hikari.UndefinedType] = False,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        mentions_reply: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialUser], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
        role_mentions: typing.Union[
            hikari.SnowflakeishSequence[hikari.PartialRole], bool, hikari.UndefinedType
        ] = hikari.UNDEFINED,
    ) -> hikari.Message:
        # <<inherited docstring from tanjun.abc.Context>>.
        delete_after = _delete_after_to_float(delete_after) if delete_after is not None else None
        async with self._response_lock:
            message = await self._message.respond(
                content=content,
                attachment=attachment,
                attachments=attachments,
                component=component,
                components=components,
                embed=embed,
                embeds=embeds,
                tts=tts,
                nonce=nonce,
                reply=reply,
                mentions_everyone=mentions_everyone,
                mentions_reply=mentions_reply,
                user_mentions=user_mentions,
                role_mentions=role_mentions,
            )
            self._last_response_id = message.id
            if self._initial_response_id is None:
                self._initial_response_id = message.id

            if delete_after is not None:
                asyncio.create_task(self._delete_after(delete_after, message))

            return message
